# (C) 2012-2015 Elke Schaper

"""
    :synopsis: Viterbi algorithm for circular hidden Markov models.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

from collections import defaultdict
import logging
import numpy as np
import operator
import re

from tral import configuration

LOG = logging.getLogger(__name__)
CONFIG = configuration.Configuration.instance().config


def viterbi(hmm, emission):
    """ Calculate the most probable sequence of states given a sequence of
    emissions and a HMM using the Viterbi algorithm

    Get local copies of all variables.
    All probabilities must be given as logarithms
    Replace Selenocysteine (U) with Cysteine (C), replace Pyrrolysine (O) with
    Lysine (K) [Reason: Seldom AAs that are not part of standard amino acid
    substitution models.]

    Args:
        hmm (hmm): An instance of the HMM class.
        emission (sequence): An instance of the Sequence class.

    Returns:
        The most likely sequence of hmm states to emit the sequence in the
        form of a list of str.

    .. todo:: Adapt docstrings to refactored Viterbi -> Viterbi_path classes.
    .. todo:: Check: Do you need local copies of all variables?
    .. todo:: Do the functions related to viterbi need to be summarized (e.g.
              in one class?) How do they relate to the Sequence class, or the
              HMM class?
    """

    if len(emission) / hmm.l_effective < \
        float(CONFIG['filter']['basic']['dict']['n_effective']['threshold']):
        LOG.info("Skip the HMM as it is too long ({}) for this sequence ({}) " \
                 "according to the filter criterion min n_effective ({})." \
                 .format(hmm.l_effective, len(emission),
                         CONFIG['filter']['basic']['dict']['n_effective']))
        return None
    if hmm.l_effective > float(CONFIG['hmm']['l_effective_max']):
        LOG.info("Skip the HMM as it is too long (%d) according to the " \
                 "filter criterion max hmm.l_effective (%d).", hmm.l_effective,
                 CONFIG['hmm']['l_effective_max'])
        return None

    states = hmm.states
    p_0 = {iS: value for iS, value in hmm.p_0.items()}
    p_e = {
        iS: {
            iE: value for iE,
            value in emission.items()} for iS,
        emission in hmm.p_e.items()}
    p_t = {
        iS: {
            iT: value for iT,
            value in transition.items()} for iS,
        transition in hmm.p_t.items()}

    if any([(iS not in CONFIG['all_amino_acids']) for iS in emission]):
        raise Exception(
            "There is an unknown amino acid in:\n {}\n".format(emission))

    # In case there are ambiguous amino acids in the sequence, calculate the
    # expected frequencies of the AAs that they could stand for from the
    # emission frequencies of the hmm in the neutral state "N".
    d_ambiguous_local = {}
    for iA in CONFIG['ambiguous_amino_acid'].keys():
        d_ambiguous_local[iA] = {}
        if iA in emission:
            total = np.log10(sum(10 ** p_e['N'][i_ambiguous] \
                    for i_ambiguous in CONFIG['ambiguous_amino_acid'][iA]))
            for i_ambiguous in CONFIG['ambiguous_amino_acid'][iA]:
                d_ambiguous_local[iA][i_ambiguous] = p_e['N'][i_ambiguous] - total

    LOG.debug("p_0: {0}".format(p_0))
    LOG.debug("p_e: {0}".format(p_e))
    LOG.debug("p_t: {0}".format(p_t))

    # Initialisation of the probabilities on the first emitted character
    if emission[0] in d_ambiguous_local:
        path = {iS: {'probability': p_0[iS], 'path': [iS]} for iS in states}
        for iS in states:
            # Calculate the average emission probability of ambiguity chars.
            # The numerical trick to calculate the average log values in high
            # precision is reused in the next for-loop and described there.
            l_p_emission = [
                d_ambiguous_local[
                    emission[0]][i_ambiguous] +
                p_e[iS][i_ambiguous] for i_ambiguous in d_ambiguous_local[
                    emission[0]]]
            max_p = max(l_p_emission)
            path[iS]['probability'] += \
               np.log10(sum([10 ** (i - max_p) for i in l_p_emission])) + max_p

    else:
        path = {iS: {'probability': p_0[iS] + p_e[iS][emission[0]],
                     'path': [iS]} for iS in states}
    LOG.debug("Path: {0}".format(path))

    # Viterbi on all remaining emitted characters
    for iE in emission[1:]:
        LOG.debug("Emitted: {0}".format(iE))

        # Determine next most probable state and its probability ...
        for iS in states:
            p = {}
            for i_former in states:
                if iE in d_ambiguous_local:
                    # Calculate the probability of being in state iS and
                    # emitting i_ambiguous for any of the AAs that the
                    # ambiguous iE stands for.
                    # Then, average over these probabilities (taking into
                    # account the background frequencies of all i_ambiguous)
                    d_p_former = {
                        i_ambiguous: probability_of_the_former_state(
                            i_former,
                            iS,
                            i_ambiguous,
                            p_e,
                            p_t,
                            path,
                            ) for i_ambiguous in d_ambiguous_local[iE].keys()}
                    d_p_former = {
                        i_ambiguous: j for i_ambiguous,
                        j in d_p_former.items() if j}
                    if len(d_p_former) > 0:
                        # Next, we need to calculated a weighted average over
                        # log10 probabilities.
                        # For this purpose, we need to transform the log10ps
                        # back to ps.
                        # However, the ps might have too small values.
                        # We apply a numerical trick:
                        # Instead of
                        # sum(w(i)p(i)) = sum(10 ** (log(w(i)) + log(p(i))))
                        # we calculate
                        # sum(w(i)p(i)) = sum(10 ** (log(w(i)) + log(p(i)) + k )) / (10 ** k),
                        # which can easily be shown to be equivalent.
                        # For (-k), we choose the maximum value in log(w(i)) +
                        # log(p(i)), as it corresponds to the maximum and thus
                        # most pronounced probability of the sum.
                        l_p_former = [
                            i +
                            d_ambiguous_local[iE][i_ambiguous] for i_ambiguous,
                            i in d_p_former.items()]
                        max_p = max(l_p_former)
                        p[i_former] = np.log10(
                            sum([10 ** (i - max_p) for i in l_p_former])) + max_p
                else:
                    p_former = probability_of_the_former_state(
                        i_former,
                        iS,
                        iE,
                        p_e,
                        p_t,
                        path)
                    if p_former:
                        p[i_former] = p_former

            # This error will occur when it is not possible to enter a state iS
            # at emission iE. This can happen for more complex HMMs, but should
            # not happen for the simple
            # circular HMMs that we consider here. It means that state iS
            # cannot be reached by emission iE. Either iS cannot emit iE, or iS
            # cannot be reached by transitions at this point.
            if p == {}:
                logging.error("The dict p is empty. This should never happen.")
            # Detect the most likely path.
            next_state, p_next_state = max(
                p.items(), key=operator.itemgetter(1))
            path[iS]['probability_next'] = p_next_state
            path[iS]['path_next'] = path[next_state]['path'][:]
            path[iS]['path_next'].append(iS)

        # ... and update the path accordingly
        for iV in path.values():
            iV['path'] = iV['path_next'][:]
            iV['probability'] = iV['probability_next']
        #LOG.debug("Path: {0}".format(path))

    # Which overall path is the most likely?
    path_summary = {iS: path[iS]['probability'] for iS in states}
    most_likely_terminal_state, p_most_likely_path = max(
        path_summary.items(), key=operator.itemgetter(1))
    most_likely_path = path[most_likely_terminal_state]['path']
    LOG.debug("The most likely path is %s. It has a score of %d.",
              str(most_likely_path), p_most_likely_path)

    return most_likely_path


def probability_of_the_former_state(i_former, iS, iE, p_e, p_t, path):
    ''' Calculate the probability of i_former, given iS and iE together with
    the dicts of emission and transition probabilities.'''
    # Exclude paths with a zero probability (i.e. a log(probability) set to
    # None)
    if iS in p_t[i_former] and None is not p_t[i_former][iS] and iE in p_e[
            iS] and None is not p_e[iS][iE] and None is not path[i_former]['probability']:
        return path[i_former]['probability'] + p_t[i_former][iS] + p_e[iS][iE]

    return None


def distance_index(i, j, length):
    """ Helper function to calculate the distance between two indices in a
    circular HMM.

    Helper function to calculate the distance between two indices in a
    circular HMM.

    Args:
        i (int): first index
        j (int): second index
        length (int): length of the HMM

    Returns:
        int: The distance between the indices ``i`` and ``j``: As the HMM
        is circular ``j`` may have a smaller value than ``i``, even though
        ``j`` is ahead of ``i``.

    .. todo:: May be replaced by a simple mod, as distance_index is used only
        once in the code at current.
    """
    if j > i:
        return j - i
    else:
        return length + j - i


def hmm_path_to_maximal_complete_tandem_repeat_units(
        sequences,
        paths,
        l_effective,
        alpha=None):
    """ Convert several viterbi paths of a hmm on several sequences into the
    corresponding hmm units.

    Be ungreedy: Start from the last index in the cluster of all start state
    and end state indices.

    Only integrate repeat units that are at least alpha complete (be it before
    of after)
    If you prefer absolute number of characters to filter which repeat units
    are used, and which not, change this in two occurrences of alpha.

    Assume that all states are counted starting on 1.

    Args:
        sequences (list of str): A list of sequences.
        paths (list of list of str): A list of Viterbi paths
        l_effective (int): length of the HMM used to create the Viterbi paths.
        alpha (Float): alpha element [0,1].

    Returns:
        A list of multiple sequence alignments (MSAs) in the form of a list of
        list of str, e.g. ``[['ATAILC', 'ATAILC', 'ACALKG'], ...]``.

    Raises:
        Exception: If alpha is not in [0,1].

    .. todo:: Use sequence instances instead of just strings for `sequences`?
    .. todo:: Check example for returns.

    """

    if not alpha:
        alpha = 0.6

    # Find the first match state index used by all paths (add None if None)
    l_terminal_indices = []
    l_repeat_present = []
    for path in paths:
        for iP in path:
            if iP.startswith("M"):
                l_terminal_indices.append(int(iP[1:]))
                l_repeat_present.append(True)
                break
        else:
            l_repeat_present.append(False)

        for iP in path[::-1]:
            if iP.startswith("M"):
                l_terminal_indices.append((int(iP[1:])) % l_effective + 1)
                break

    # Find the max distance between two indices
    l_used_indices = sorted(
        [i for i in set(l_terminal_indices) if i is not None])

    if len(l_used_indices) == 0:
        return None
    elif len(l_used_indices) == 1:
        start_index = l_used_indices[0]
    else:
        distances = [i - j for j, i in \
                        zip(l_used_indices[:-1], l_used_indices[1:])] + \
                    [l_effective - l_used_indices[-1] + l_used_indices[0]]
        max_distance_index, max_distance = max(enumerate(distances),
                                               key=operator.itemgetter(1))

        # Define the start_index
        start_index = l_used_indices[
            (max_distance_index + 1) %
            len(l_used_indices)]

    # Get all TR units according to shift.
    lMSA = []

    for ib_present, i_seq, i_path in zip(l_repeat_present, sequences, paths):

        if not ib_present:
            lMSA.append(None)
            continue

        msa = []
        # Intialise the last used index and the first msa_unit
        current_index = 1
        msa_unit = ''
        for iS, iP in zip(i_seq, i_path):
            if iP == 'N':
                continue
            elif iP == 'C':
                break
            match_state_index = (int(iP[1:]) + (l_effective - start_index)) % \
                                                l_effective + 1
            if match_state_index >= current_index:
                # We are staying within the same repeat unit.
                msa_unit += iS
            else:
                # We are starting a new repeat unit.
                msa.append(msa_unit)
                msa_unit = iS
            current_index = match_state_index

        msa.append(msa_unit)

        try:
            if len(msa[-1]) < alpha * l_effective:
                msa = msa[:-1]
            if len(msa[0]) < alpha * l_effective:
                msa = msa[1:]
            lMSA.append(msa)
        except:
            lMSA.append([])

    return lMSA


def hmm_path_to_non_aligned_tandem_repeat_units(sequence, path, l_effective):
    """ Convert a viterbi <path> of a hmm of length <l_effective> on <sequence>
     into the corresponding tandem repeat

    Extract the tandem repeat alignment from a sequence given a Viterbi path.
    Ignore the alignment information in the Viterbi path. For example, all
    emissions labelled with M1 (match state 1) align according to the HMM.
    However, this method does not use this information. Therefore, for example
    the first characters in the repeat units do not necessarily align, and the
    repeat units are not necessarily of same length.

    Assume that all states are counted starting on 1.

    Args:
        sequence (str): A sequence as a string.
        paths (list of list of str): A list of Viterbi paths
        l_effective (int): length of the HMM used to create the Viterbi paths.

    Returns:
        A multiple sequence alignment (MSA) created from the most likely path
        along the hmm in the form of a list of str.

    .. todo:: Use sequence instance instead of string for input?

    """

    begin = path.count('N')
    # If no repeat was found, return None.
    if begin == len(sequence) or path.count('C') == len(sequence):
        return None

    splitter = re.compile("(\w)(\d+)")
    mapping = [
        ((splitter.match(iP).group(1), int(splitter.match(iP).group(2))), iS) \
        for iS, iP in zip(sequence[begin:], path[begin:]) if iP != 'C']

    shift = mapping[0][0][1]
    index_shift = ["Empty"] + [(i + l_effective + 1 - shift) % \
                              l_effective for i in range(l_effective)]
    LOG.debug("The tandem repeat is shifted by: %d", shift)

    repeat_msa = []
    repeat_unit = mapping[0][1]
    last_used_index = 0

    if l_effective == 1:
        for iM, iS in mapping[1:]:
            if iM[0] == "I":
                repeat_unit += iS
            else:
                repeat_msa.append(repeat_unit)
                repeat_unit = iS

    else:
        for iM, iS in mapping[1:]:
            #print("{} {}".format(iM,iS))
            if index_shift[iM[1]] >= last_used_index:
                repeat_unit += iS
            else:
                repeat_msa.append(repeat_unit)
                repeat_unit = iS
            last_used_index = index_shift[iM[1]]

    repeat_msa.append(repeat_unit)

    return repeat_msa


def hmm_path_to_aligned_tandem_repeat_units(sequence, most_likely_path, l_effective,
                                            translate=False):
    """Convert a viterbi path in an hmm of length ``l_effective`` on the
     sequence into a corresponding tandem repeat.

    Extract the tandem repeat alignment from a sequence given a Viterbi path.
    Use alignment information in the Viterbi path. For example, all emissions
    labelled with M1 (match state 1) align according to the HMM. Insert gaps
    for insertions and deletions accordingly.
    Thus, for example the first characters in the repeat units do necessarily
    align, albeit some of them may be gaps. Also, all repeat units are
    necessarily of same length.

    Assume that all states are counted starting on 0, unless the ``translate``
    flag is set.

    Args:
        sequence (str): A sequence as a string.
        most_likely_path (list of str): a Viterbi path.
        l_effective (int): length of the HMM used to create the Viterbi paths.
        translate (bool): This function assumes that HMM states are enumerated
            starting on 0.
            If the HMM states are enumerated starting on 1, set this flag for
            transformation.

    Returns:
        The function returns a tuple consisting of 3 values.
        The tuple contains:

        * ``msa``: A repeat instance.
        * ``begin``: The start index of the repeat on the sequence.
        * ``shift``: The index of the HMM where the cut between repeat units is set.


    .. warning:: [None].
    .. todo:: Use sequence instance instead of just a string?
    .. todo:: Check: How is the returned `begin` defined? Starting counting on 0 or 1?
        Is it the index of the last flanking character, or the first repeat character?
    .. todo:: Can we update this function, e.g. to not assume that HMM states start on 0?
    .. todo:: Check the docstring, reformat returns.
    """

    begin = most_likely_path.count('N')
    # If no repeat was found, return None.
    if begin == len(sequence) or most_likely_path.count('C') == len(sequence):
        return None

    if translate:
        # In the input data, the states are counted starting on 1. Subtract 1.
        most_likely_path = [
            i if i in ["C", "N"] else i[0] + str(int(i[1:]) - 1) for i in most_likely_path]

    splitter = re.compile("(\w)(\d+)")

    mapping = [
        ((splitter.match(iP).group(1), int(splitter.match(iP).group(2))), iS) \
        for iS, iP in zip(sequence[begin:], most_likely_path[begin:]) \
        if iP != 'C']

    shift = mapping[0][0][1]
    index_shift = [(i + l_effective - shift) % l_effective for i in range(l_effective)]
    LOG.debug("The tandem repeat is shifted by: %d", shift)

    insertions = []
    repeat_text = mapping[0][1]
    insertions.append(defaultdict(str))
    max_used_index_M = 0
    max_used_index_I = -1
    last_used_index = shift

    for iM, iS in mapping[1:]:

        LOG.debug("Iteration iS: %d, iM: %d, last_used_index: %d, " \
                  "max_used_index_M: %d, max_used_index_I: %d",
                  iS, iM, last_used_index, max_used_index_M, max_used_index_I)

        # If we have entered a new repeat unit, add a new element to <insertions>
        # (Including some index magic, e.g. the insertion state index is shifted by one (lowered))
        if iM[0] == "M" and (index_shift[iM[1]] <= max_used_index_M or \
                             index_shift[iM[1]] <= max_used_index_I):
            insertions.append(defaultdict(str))
            max_used_index_M = index_shift[iM[1]]
            max_used_index_I = index_shift[iM[1]] - 1

        elif iM[0] == "I" and (index_shift[iM[1] - 1] < max_used_index_I or \
                              index_shift[iM[1] - 1] < max_used_index_M):
            insertions.append(defaultdict(str))
            max_used_index_I = index_shift[iM[1] - 1]
            max_used_index_M = index_shift[iM[1] - 1]

        if iM[0] == "M":
            max_used_index_M = index_shift[iM[1]]
        else:
            max_used_index_I = index_shift[iM[1] - 1]

        # Save match state and deletion information
        if iM[0] == "M":
            n_deletions = distance_index(last_used_index, iM[1], l_effective) - 1
            repeat_text += "-" * n_deletions + iS
            last_used_index = iM[1]

        else:
            insertions[-1][index_shift[iM[1]]] += iS

    msa = [repeat_text[i:i + l_effective] for i in range(0, len(repeat_text), l_effective)]
    msa[-1] += "-" * (l_effective - len(msa[-1]))
    msaT = ["".join(c) for c in zip(*msa)]
    n = len(msa)
    LOG.debug("This tandem repeat has %d repeat units.", n)
    LOG.debug("These insertions were detected: %s.", str(insertions))

    msa_temp = []
    # For each site ...
    for i in range(l_effective):
        msa_temp.extend(msaT[i:i + 1])
        # ... for each tandem repeat unit, check whether there are insertions.
        for j, row in enumerate(insertions):
            # Backtranslate the index used in insertions (shift by one to the
            # right (increase))
            if (i + 1) % l_effective in row:
                for iS in row[(i + 1) % l_effective]:
                    msa_temp.append("-" * j + iS + "-" * (n - j - 1))
    msa = ["".join(c) for c in zip(*msa_temp)]

    return msa, begin, shift



import os
import pytest

from tral.paths import config_file

from tral.sequence import repeat_detection_run, sequence

TEST_SEQUENCE_Q9BRR0 = "MARELSESTALDAQSTEDQMELLVIKVEEEEAGFPSSPDLGSEGSRERFRGFRYPEAAGPREALSRLRELCRQWLQPEMHSKEQILELLVLEQFLTILPGNLQSWVREQHPESGEEVVVLLEYLERQLDEPAPQVSGVDQGQELLCCKMALLTPAPGSQSSQFQLMKALLKHESVGSQPLQDRVLQVPVLAHGGCCREDKVVASRLTPESQGLLKVEDVALTLTPEWTQQDSSQGNLCRDEKQENHGSLVSLGDEKQTKSRDLPPAEELPEKEHGKISCHLREDIAQIPTCAEAGEQEGRLQRKQKNATGGRRHICHECGKSFAQSSGLSKHRRIHTGEKPYECEECGKAFIGSSALVIHQRVHTGEKPYECEECGKAFSHSSDLIKHQRTHTGEKPYECDDCGKTFSQSCSLLEHHRIHTGEKPYQCSMCGKAFRRSSHLLRHQRIHTGDKNVQEPEQGEAWKSRMESQLENVETPMSYKCNECERSFTQNTGLIEHQKIHTGEKPYQCNACGKGFTRISYLVQHQRSHVGKNILSQ"

#@pytest.mark.slow
def test_detect_TRUST():

    test_seq = sequence.Sequence(TEST_SEQUENCE_Q9BRR0)
    predicted_repeats = repeat_detection_run.run_detector(seq_records = [test_seq], detectors = ["TRUST"])[0]['TRUST']
    # Warning: TRUST finds these results ONLY with the BLOSUM50 substitution matrix.
    assert len(predicted_repeats) == 3
    # Warning: TRUST finds these results ONLY with the BLOSUM50 substitution matrix.
    assert predicted_repeats[0].msa == ['HLREDIAQIP---TCAEAGE---QEGRLQR', 'KQKNATGGRR--HICHECGKSFAQSSGLSK', 'HRRIHTGEKP--YECEECGKAFIGSSALVI', 'HQRVHTGEKP--YECEECGKAFSHSSDLIK', 'HQRTHTGEKP--YECDDCGKTFSQSCSLLE', 'HHRIHTGEKP--YQCSMCGKAFRRSSHLLR', 'HQRIHTGDKN--VQEPEQGEAW--KSRM--', 'ESQLENVETPmsYKCNECERSFTQNTGLIE', 'HQKIHTGEKP--YQCNACGKGFTRISYLVQ']


def test_detect_XSTREAM():
    test_seq = sequence.Sequence(TEST_SEQUENCE_Q9BRR0)
    predicted_repeats = repeat_detection_run.run_detector(seq_records = [test_seq], detectors = ["XSTREAM"])[0]['XSTREAM']
    assert len(predicted_repeats) == 1
    assert predicted_repeats[0].msa == ['ECGKSFAQS-SGLSK-HRRIHTGEKPYECE', 'ECGKAFIGS-SALVI-HQRVHTGEKPYECE', 'ECGKAFSHS-SDL-IKHQRTHTGEKPYECD', 'DCGKTFSQSCSLLEH-H-RIHTGEKPY']


def test_detect_TREKS():

    test_seq = sequence.Sequence(TEST_SEQUENCE_Q9BRR0)
    predicted_repeats = repeat_detection_run.run_detector(seq_records = [test_seq], detectors = ["T-REKS"])[0]['T-REKS']
    assert len(predicted_repeats) == 1
    assert predicted_repeats[0].msa == ['C---G---KSFAQSSGLSKHRRIHTGEKPYECE-E', 'C---G---KAFIGSSALVIHQRVHTGEKPYECE-E', 'C---G---KAFSHSSDLIKHQRTHTGEKPYECD-D', 'C---G---KTFSQSCSLLEHHRIHTGEKPYQCS-M', 'C---G---KAFRRSSHLLRHQRIHTGDKNVQ-EPE', 'Q---G---EAW--KSRME-SQ-LENVETPMSYK--', 'C---NECERSFTQNTGLIEHQKIHTGEKPYQ----', 'CNACG---KGFTRISYLVQHQRSHVG-KNI-LS--']


def test_detect_HHrepID():

    test_seq = sequence.Sequence(TEST_SEQUENCE_Q9BRR0)
    predicted_repeats = repeat_detection_run.run_detector(seq_records = [test_seq], detectors = ["HHrepID"])[0]['HHrepID']
    assert len(predicted_repeats) == 1
    assert predicted_repeats[0].msa == ['------------------IPTCAEAGEQ----', 'EGRLQRKQKNATGGRRHICHECGKSFAQ----', 'SSGLSKHRRIHTGEKPYECEECGKAFIG----', 'SSALVIHQRVHTGEKPYECEECGKAFSH----', 'SSDLIKHQRTHTGEKPYECDDCGKTFSQ----', 'SCSLLEHHRIHTGEKPYQCSMCGKAFRR----', 'SSHLLRHQRIHTGDKNVQEPEQGEAWKSRMES', '------QLENVETPMSYKCNECERSFTQ----', 'NTGLIEHQKIHTGEKPYQCNACGKGFTR----', 'ISYLVQHQRSHVG-------------------']



# (C) 2011, Alexander Korsunsky
# (C) 2011-2015 Elke Schaper

"""
    :synopsis: Execution of repeat detection algorithms

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

from collections import OrderedDict
import distutils
import itertools
import logging
import os
import re
import resource
import shutil
import subprocess
import sys
import tempfile

from tral import configuration
from tral.sequence import repeat_detection_io

LOG = logging.getLogger(__name__)

CONFIG_GENERAL = configuration.Configuration.instance().config
REPEAT_DETECTOR_PATH = CONFIG_GENERAL["sequence"]["repeat_detector_path"]


class BinaryExecutable:

    """ Contains the executable, and combines executable with parameters.

    Contains the executable, and combines executable with parameters.

     Attributes:
        binary(str): Path to binary

    """

    def __init__(self, binary=None, name=None):
        """ Construct a BinaryExecutable object.

        Construct a BinaryExecutable object.

        Args:
            binary (str): Path to the binary
            name(str): Name of the programme.
        """

        if not binary:
            raise TypeError("A binary executable must be provided.")
        try:
            self.binary = shutil.which(binary)
        except:
            self.binary = distutils.spawn.find_executable(binary)
        if not self.binary:
            raise ValueError(
                "The executable {} does not exist, although {} was selected "
                "to be executed. Please make sure the executable is in the system path, or "
                "the path to the executable is correctly set in config.ini".format(
                    binary, name))

    def get_execute_tokens(self, *args):
        """Return the tokens to invoke the program with the arguments args"""

        return [self.binary] + list(args)

    def get_execute_line(self, *args):
        """Return the command line to invoke the program with the arguments args"""
        return " ".join(self.get_execute_tokens(*args))


def check_java_errors(outfile, errfile, log=LOG, procname=None):
    """ Check for java problems. Return True if there were problems, else False.

    Check for these java errors:

     * Stdout file is empty but stderr file is not.
     * Java Exception string is indicated in the errfile

     Return True if there were problems, else False.

    Args:
        outfile (file handle):  Redirected standard output channel file.
        errfile (file handle): Redirected standard error channel file
            If None, no copies are saved.
        log (?): Name of the log to issue log messages to. If none, no log
                messages will be issued.

    .. todo:: Complete docstring
    """

    has_error = False

    with open(outfile, mode='r') as of, open(errfile, mode='r') as ef:
        outfile_size = of.seek(0, os.SEEK_END)
        errfile_str = ef.read()

    errfile_size = len(errfile_str)

    if errfile_size != 0 and outfile_size == 0:
        has_error = True
        LOG.warning(
            "Process \"%s\" has empty STDOUT but non-empty STDERR",
            procname)

    pat_javaexc = re.compile(
        r'Exception in thread ".+" \S+:.*$(\s+at .+)+',
        re.M)

    m = pat_javaexc.search(errfile_str)
    if m:
        has_error = True
        LOG.warning(
            "Java Exception probably occured in process \"%s\"! Exception information:\n%s",
            procname,
            m.group(0))

    return has_error


class TRDetector(object):

    def __init__(self, executable):
        """Construct a TRDetector object with executable als executable object"""
        LOG.debug(executable)
        self.__executable = executable

    def run_process(self, working_dir, *args):
        """Launch a detector process

        Arguments:
        working_dir -- Working directory to run process in
        args -- Arguments to pass to the process

        Return Value:
        A tuple with:
            - The return code
            - The name to the file with the redirected standard output channel (stdout)
            - The name to the file with the redirected standard error channel (stderror)

        The standard output (stdout) and standard error (stderr) channels will
        be redirected to working_dir/stdout.txt and working_dir/stdout.txt
        respectively.
        """

        # create working directory in top level working directory,
        # if it does not yet exist
        if not os.path.isdir(working_dir):
            os.makedirs(working_dir)

        stdoutfname = os.path.join(working_dir, "stdout.txt")
        stderrfname = os.path.join(working_dir, "stderr.txt")

        # open stdout and stderr output file
        __stdout_file = open(stdoutfname, mode='w')
        __stderr_file = open(stderrfname, mode='w')

        LOG.debug("Launching process: %s in %s",
                  self.__executable.get_execute_line(*args), working_dir)

        LOG.debug("Launching process tokens: %s in %s",
                  self.__executable.get_execute_tokens(*args), working_dir)
        # launch process
        __process = subprocess.Popen(
            self.__executable.get_execute_tokens(*args), cwd=working_dir,
            stdout=__stdout_file, stderr=__stderr_file, close_fds=True,
        )
        __process.wait()

        __stdout_file.close()
        __stderr_file.close()

        return __process.returncode, stdoutfname, stderrfname


class DetectorHHrepID(TRDetector):
    name = 'HHrepID'
    displayname = "HHrepID"

    """ Execute vi
    ./hhrepid_32 -i infile -v 0 -nofilt -d dummyHMM.hmm -o resultfile
    ./hhrepid -i <query> -d <path to cal.hhm> -tp <path to tp.dat> -fp <path to fp.dat> """

    class Configuration:

        def __init__(self):
            self.boolopts = {
                # turn off low-complexity filter (default: filter on)
                "-nofilt": False,
                # calibration by shuffling instead of database search (default:
                # off)
                "-shuffle": False
            }

            self.valopts = {
                # <file> input query alignment  (fasta/a2m/a3m) or HMM file (.hhm)
                "-i": None,
                "-d": REPEAT_DETECTOR_PATH['HHrepID_dummyhmm'],   # <path> dummy hmm database file
                "-o": 'hhrepID.o',    # <file> write results and multiple sequence alignment to file (default=none)
                "-v": 0,           # -v: verbose mode (default: show only warnings)  ;  -v 0: suppress all screen outpu
                "-P": None,        # <float> max p-value of suboptimal alignments in all search rounds but the last one (def=0.1)
                "-R": None,        # <float> max p-value of repeats  (def=1)
                "-T": None,        # <float> max total repeat p-value (def=0.001)
                "-alpha": None,        # <float> For calculating repeat p-values: weight of n'th suboptimal alignment vs. (n+1)-st suboptimal alignmen
                "-k": None,        # [0,inf[ For calculating repeat p-values: maximal number of suboptimal alignments considered (def=1)
                "-cont": None,        # <float>  probability threshold for masking of inconsistent cells (def=0.001)
                "-mrgr": None,        # [0,inf[  number of merge rounds for achieving consistency (def=3)
                "-lcon": None,        # <float>  preserved local context in shuffling (def=2)
                "-lmin": None,        # [0,inf[  minimal length of repeats to be identified (def=7)

            }

        def tokens(self, infile=None):
            """Generate command line tokens based on this configuration.
            Arguments:

            infile -- generate tokens to pass infile as input file to detector"""
            toks = []
            if infile:
                toks += ['-i', infile]

            for optstring, optvalue in self.valopts.items():
                if optvalue is not None:
                    toks += [str(optstring), str(optvalue)]

            for optstring, optvalue in self.boolopts.items():
                if optvalue:
                    toks.append(optstring)

            return toks

    def __init__(self, name=name):
        """Construct DetectorHHrepID object.

        Arguments:
        name: The name of the tandem repeat detector.
        """

        self.config = self.Configuration()
        executable = BinaryExecutable(
            binary=REPEAT_DETECTOR_PATH[name],
            name=name)
        super(DetectorHHrepID, self).__init__(executable)

    def run_process(self, working_dir, infile):
        """Run detector process on infile in working_dir and return all repeats found"""

        wd = os.path.join(working_dir, DetectorHHrepID.name)

        prog_args = self.config.tokens(infile=infile)

        # execute detector process
        retcode, stdoutfname, stderrfname = \
            super().run_process(wd, *prog_args)

        # CHECK WHETHER outfile exists, else give warning and return empty
        # list.

        # Process output file, return results
        resultfile = os.path.join(wd, self.config.valopts["-o"])
        if os.path.isfile(resultfile):
            with open(resultfile, "r") as infilehandle:
                tmp = list(
                    repeat_detection_io.hhpredid_get_repeats(infilehandle))
            return tmp
        else:
            return []


class DetectorPhobos(TRDetector):
    name = 'PHOBOS'
    displayname = "PHOBOS"

    # execute via phobos --printRepeatSeqMode 3 DNA.faa phobos.o
    # or phobos --help to see all options

    class Configuration:

        def __init__(self):
            self.boolopts = {
                # Phobos will choose the shorter of the two alignments instead
                # of the longer. Phobos will run faster with this option.
                "--preferShorterRepeats": False,
                # dontRemoveMostlyOverlapping. Default: DO remove mostly
                # overlapping repeats in favour of the one with the higher
                # internal score
                "--dontRemoveMostlyOverlapping": False,
            }

            self.scriptopts = {
                "outputfile": 'phobos.o',
            }

            self.valopts = {
                # <int> Display of TR. 0: asIs, 1: Alphabetical normal form, 2: Alphabetical normal form also considering the reverse complement. Default: 2
                "--reportUnit": None,
                "--NPerfectionMode": None,  # <int> Treatment of N.  0: asMismatch, 1: asNeutral, 2: asMatch. Default: 0.
                "--printRepeatSeqMode": '3',  # <int> Show the repeat unit alignment
                "--outputFormat": None,  # <int> 0 : Use the Phobos output format. Default: 0
                "--minPerfection": None,  # <float> Minimum perfection of a satellites. Default: 0.
                "--maxPerfection": None,  # <float> Maximum perfection of a satellites. Default: 100.
                "--maximum_score_reduction": None,  # The maximum amount the score can be reduced before search is aborded. Typical: 6*mismatch-penalty or infinite. Default: infinite
                "--recursion": None,  # The recursion depth used in the search. Values in the range 3 to 7 are recommended. A value of 0 implies a search for perfect repeats only. Default: Not stated
                "--minUnitLen": None,  # <int> Minimum unit length. Default: 1
                "--maxUnitLen": None,  # <int> Maximum unit length. Default: 10
                "--minLength_b": None,  # <float> The minimum length of a repeat is determined with: maximum( minLength, minLength_a + minLength_b*(unit-length) ). Default value of minLength_b: 0
                "--minLength_a": None,  # <int> Default value of minLength_a: 0
                "--minLength": None,  # <int> Default value of minLength: 0
                "--minScore_b": None,  # <float> The minimum score of a repeat is determined with: maximum( minScore, minScore_a + minScore_b*(unit-length) ). Default value of minScore_b: 1
                "--minScore_a": None,  # <int> Default value of minScore_a: 0
                "--minScore": None,  # <int> Default value of minScore: 6
                "--mismatchScore": None,  # <int> Score for mismatch - must be negative. Default: -5. Match score is fixed to one.
                "--indelScore": None,  # <int> Score for indels - must be negative. Default: -5. Match score is fixed to one.
                "--searchMode": 'imperfect'  # ['exact','extendExact','imperfect']
            }

        def set_working_dir(self, working_dir):
            self.scriptopts['outputfile'] = os.path.join(
                working_dir,
                self.scriptopts['outputfile'])
            if not os.path.isdir(working_dir):
                os.makedirs(working_dir)

        def tokens(self, infile=None):
            """Generate command line tokens based on this configuration.
            Arguments:

            infile -- generate tokens to pass infile as input file to detector"""
            toks = [
                optstring for optstring,
                optvalue in self.boolopts.items() if optvalue]

            for optstring, optvalue in self.valopts.items():
                if optvalue is not None:
                    toks.append(optstring)
                    toks.append(str(optvalue))

            if infile:
                toks.append(infile)
            toks.append(self.scriptopts['outputfile'])
            return toks

    def __init__(self, name=name):
        """Construct DetectorPhobos object.

        Arguments:
        name: The name of the tandem repeat detector.
        """

        self.config = self.Configuration()
        executable = BinaryExecutable(
            binary=REPEAT_DETECTOR_PATH[name],
            name=name)
        super(DetectorPhobos, self).__init__(executable)

    def run_process(self, working_dir, infile):
        """Run detector process on infile in working_dir and return all repeats found"""
        wd = os.path.join(working_dir, DetectorPhobos.name)
        self.config.set_working_dir(working_dir=wd)

        # execute detector process
        retcode, stdoutfname, stderrfname = super(
            DetectorPhobos, self).run_process(
            wd, *self.config.tokens(infile))
        # alternatively:
        #prog_args = self.config.tokens(infile=infile)
        #retcode, stdoutfname, stderrfname = super().run_process(wd, *prog_args)

        # Process output file, return results If outfile does not exist return
        # empty list
        if os.path.isfile(self.config.scriptopts['outputfile']):
            with open(self.config.scriptopts['outputfile'], "r") as resultfilehandle:
                tmp = list(
                    repeat_detection_io.phobos_get_repeats(resultfilehandle))
            return tmp
        else:
            LOG.warning(
                "Did not find Phobos result file in %s",
                self.config.scriptopts['outputfile'])
            return []


class DetectorTRED(TRDetector):
    name = 'TRED'
    displayname = "TRED"

    """ tred is a shell script executing:
            tred1 myDNA.faa intermediate_output
            tred2 myDNA.faa intermediate_output output1 output2
        At the moment, we do not know how to add parameters to TRED    """

    class Configuration:

        def __init__(self):
            # For now, use Tred with default options only
            self.boolopts = {
            }

            self.valopts = {
            }

        def set_result_file(self, result_file=None):
            self.result_file = result_file

        def tokens(self, infile=None):
            """Generate command line tokens based on this configuration.
            Arguments:

            infile -- generate tokens to pass infile as input file to detector"""
            toks = []
            if infile:
                toks = [infile]
            toks.append(self.result_file)

            return toks

    def __init__(self, name=name):
        """Construct DetectorTRED object.

        Arguments:
        name: The name of the tandem repeat detector.
        """

        self.config = self.Configuration()
        executable = BinaryExecutable(
            binary=REPEAT_DETECTOR_PATH[name],
            name=name)
        super(DetectorTRED, self).__init__(executable)

    def run_process(self, working_dir, infile):
        """Run detector process on infile in working_dir and return all repeats found"""

        wd = os.path.join(working_dir, DetectorTRED.name)

        self.result_file = os.path.join(wd, 'tred.o')

        self.config.set_result_file(result_file=self.result_file)
        prog_args = self.config.tokens(infile=infile)

        # execute detector process
        retcode, stdoutfname, stderrfname = \
            super().run_process(wd, *prog_args)

        # CHECK WHETHER outfile exists, else give warning and return empty
        # list.

        # Process output file, return results
        if os.path.isfile(self.result_file):
            with open(self.result_file, "r") as infilehandle:
                tmp = list(repeat_detection_io.tred_get_repeats(infilehandle))
            return tmp
        else:
            LOG.warning(
                "Did not find TRED result file in %s",
                self.result_file)
            return []


class DetectorTREKS(TRDetector):
    name = 'T-REKS'
    displayname = "T-REKS"

    class Configuration:

        def __init__(self):
            self.boolopts = {
                "-overlapfilter": True,
                "-nosplit": True
            }

            self.valopts = {
                "-msaMode": None,
                "-clustal": None,
                "-db": None,
                "-muscle": None,
                "-outfile": None,
                "-similarity": None,
                "-kmeans": None,
                "-varIndels": None,
                "-type": None
            }

        def tokens(self, infile=None):
            """Generate command line tokens based on this configuration.
            Arguments:

            infile -- generate tokens to pass infile as input file to detector"""
            bool_toks = [
                optstring for optstring,
                optvalue in self.boolopts.items() if optvalue]
            value_toks = ([
                optstring + "=" + str(optvalue)
                for optstring, optvalue in self.valopts.items()
                if optvalue is not None
            ])

            if infile:
                value_toks.append("-infile=" + infile)

            return bool_toks + value_toks

    def __init__(self,
                 name=name,
                 ):
        """Construct DetectorTREKS object.

        Arguments:
        name: The name of the tandem repeat detector.
        """

        self.config = self.Configuration()
        executable = BinaryExecutable(
            binary=REPEAT_DETECTOR_PATH[name],
            name=name)
        super(DetectorTREKS, self).__init__(executable)

    def run_process(self, working_dir, infile):
        """Run detector process on infile in working_dir and return all repeats found"""
        wd = os.path.join(working_dir, DetectorTREKS.name)

        prog_args = self.config.tokens(infile=infile)

        # execute detector process
        retcode, stdoutfname, stderrfname = \
            super(DetectorTREKS, self).run_process(wd, *prog_args)

        if check_java_errors(stdoutfname, stderrfname,
                             log=LOG, procname=DetectorTREKS.displayname):
            return []

        # Process output file, return results
        with open(stdoutfname, "r") as outfile:
            tmp = list(repeat_detection_io.treks_get_repeats(outfile))
        return tmp


class DetectorTRF(TRDetector):
    name = 'TRF'
    displayname = "TRF_Benson"

    # Execute via:
    # ./trf404.linux64.exe inputfile 2 7 7 80 10 50 500 -d
    # Result file in this case is called inputfile.2.7.7.80.10.50.500.1.txt.html
    # WARNING: This result file name is only used if there is only one sequence in the input file!
    ## result_file_name = ".".join([inputfile] + [str(optvalue) for optstring, optvalue in self.valopts.items() if optvalue != None] + ['1.txt.html'])

    # ./trf404.linux64.exe File Match Mismatch Delta PM PI Minscore MaxPeriod [options]

    class Configuration:

        def __init__(self):
            self.boolopts = {
                # no redundancy elimination # Already without the -r flag set,
                # the three best redundant solutions are displayed
                "-r": False,
                #"-h"            : False,    # When -h is set, the output html of interest is not written :(
                #"-d"            : False,    # produce data file, not needed at curren
                #"-m"            : False,    # masked sequence file
                #"-f"            : False,    # flanking sequence
            }

            # The order of TRF flags matters. Hence, we use an ordered
            # dictionary for valopts.
            valopts_tmp = [
                ("File", None),  # input file
                ("Match", 2),     # matching weigh
                ("Mismatch", 7),     # mismatching penalty
                ("Delta", 7),     # indel penalty
                ("PM", 80),    # integer match probability
                ("PI", 10),    # integer indel probability
                ("Minscore", 50),    # minimum alignment score to report
                ("MaxPeriod", 1000)    # maximum period size to report

            ]
            self.valopts = OrderedDict(valopts_tmp)

        def tokens(self, infile=None):
            """Generate command line tokens based on this configuration.
            Arguments:

            infile -- generate tokens to pass infile as input file to detector"""

            toks = []
            if infile:
                toks.append(infile)

            for optstring, optvalue in self.valopts.items():
                if optvalue is not None:
                    toks.append(str(optvalue))

            for optstring, optvalue in self.boolopts.items():
                if optvalue:
                    toks.append(optstring)

            return toks

    def __init__(self, name=name):
        """Construct DetectorTRF object.

        Arguments:
        name: The name of the tandem repeat detector.
        """

        self.config = self.Configuration()
        executable = BinaryExecutable(
            binary=REPEAT_DETECTOR_PATH[name],
            name=name)
        super(DetectorTRF, self).__init__(executable)

    def run_process(self, working_dir, infile):
        """Run detector process on infile in working_dir and return all repeats found"""

        wd = os.path.join(working_dir, DetectorTRF.name)

        prog_args = self.config.tokens(infile=infile)

        # execute detector process
        retcode, stdoutfname, stderrfname = \
            super().run_process(wd, *prog_args)

        # get the infile name from the infile path
        infile_name = re.findall(r"/([_\.\w]+)$", infile)[0]

        # WARNING! TRF is a volatile programme, this result_file_name might not always be true.
        # Maybe a general search on txt.html elements in the result dir might
        # be more appropiate.
        result_file_name = ".".join(
            [infile_name] +
            [
                str(optvalue) for optstring,
                optvalue in self.config.valopts.items() if optvalue is not None] +
            ['1.txt.html'])

        # Process output file, return results
        with open(os.path.join(wd, result_file_name), "r") as outfile:
            tmp = list(repeat_detection_io.trf_get_repeats(outfile))
        return tmp


class DetectorTrust(TRDetector):
    name = 'TRUST'
    displayname = "TRUST"

    class Configuration:

        def __init__(self):
            self.boolopts = {
                "-noseg": True,
                "-force": False,
            }

            self.valopts = {
                "-matrix": REPEAT_DETECTOR_PATH['TRUST_substitutionmatrix'],
                "-gapo": "8",
                "-gapx": "2",
                "-procTotal": "1",
            }  # procTotal: When running on a cluster, specify the amount of processors used  # procNr cpu_number:  When running on a cluster, specify 0-based CPU-number and amount of processors used

        def tokens(self, infile=None):
            """Generate command line tokens based on this configuration.
            Arguments:

            infile -- generate tokens to pass infile as input file to detector"""
            toks = [
                optstring for optstring,
                optvalue in self.boolopts.items() if optvalue]

            for optstring, optvalue in self.valopts.items():
                if optvalue is not None:
                    toks.append(optstring)
                    toks.append(str(optvalue))

            if infile:
                toks.append("-fasta")
                toks.append(infile)

            return toks

    def __init__(self, name=name):
        """Construct DetectorTrust object.

        Arguments:
        name: The name of the tandem repeat detector.
        """

        self.config = self.Configuration()
        executable = BinaryExecutable(
            binary=REPEAT_DETECTOR_PATH[name],
            name=name)
        super(DetectorTrust, self).__init__(executable)

    def run_process(self, working_dir, infile):
        """Run detector process on infile in working_dir and return all repeats found"""
        wd = os.path.join(working_dir, DetectorTrust.name)

        # execute detector process
        retcode, stdoutfname, stderrfname = super(
            DetectorTrust, self).run_process(
            wd, *self.config.tokens(infile))

        if check_java_errors(stdoutfname, stderrfname,
                             log=LOG, procname=DetectorTrust.displayname):
            return []

        # shutil.copyfile(stdoutfname, YOUR FAVOURITE PATH)

        # Process output file, return results
        with open(stdoutfname, "r") as outfile:
            return list(repeat_detection_io.trust_get_repeats(outfile))


class DetectorXStream(TRDetector):
    name = 'XSTREAM'
    displayname = "XSTREAM"

    class Configuration:

        def __init__(self):
            self.boolopts = {
                #"-n" : False,   # sequence type: nucleotide
                "-l": False,   # don't look for periods >1000
                #"-p" : False,   # don't print repeat alignmen
                "-h": True,    # write to console, not HTML
                #"-C" : False,   # don't print any repeat info
                "-z": True,  # generate spreadsheet .xls output
                #"-Z" : False,   # create statistics spreadshee
                #"-s" : False,   # parse output by input sequence
                "-f": False,   # perform divide & conquer on inpu
                #"-G" : False,   # create color-coded TR block diagram
                #"-G*" : False,  # create single color TR block diagram
                #"-B" : False,   # create sequence comparison map PNG
                #"-O" : False,   # print "mismatch/gaps colored" HTML outpu
                "-N": False,   # turn off nesting
                "-o": False,   # don't remove overlapping TR domain
                #"-S" : False,   # input strain into database
                "-V": True,   # don't invoke D & C automatically for long inpu
            }

            self.valopts = {
                "-g": None,    # [integer] set gap number
                "-i": None,    # [0-1] set matching threshold for detection
                "-I": None,    # [0-1] set cons error threshold for printing
                "-D": None,    # [0-1] set max % indels for printing
                "-m": 1,    # [integer] set minimum period
                "-x": None,    # [integer] set maximum period
                "-e": None,    # [any number>=2] set minimum copy number
                # "-a" : None,    # [string] add identifier string to output file
                "-f": None,    # [integer] set fragment length
                #"-d" : None,    # [file path] specify output directory
                #"-A" : None,    # [file] import substitution alphabe
                "-L": 6,    # [integer] set minimum TR domain length
                "-P": 0,    # [0-1] set minimum % sequence coverage
                #"-Q" : None,    # [user,pass,db] send output to MySQL TR database
            }

        def tokens(self, infile=None):
            """Generate command line tokens based on this configuration.
            Arguments:

            infile -- generate tokens to pass infile as input file to detector"""
            toks = [
                optstring for optstring,
                optvalue in self.boolopts.items() if optvalue]

            toks = toks + \
                [optstring + str(optvalue)
                    for optstring, optvalue in self.valopts.items()
                    if optvalue is not None]

            if infile:
                toks.append(infile)

            return toks

    def __init__(self, name=name):
        """Construct DetectorXStream object.

        Arguments:
        name: The name of the tandem repeat detector.
        """

        self.config = self.Configuration()
        executable = BinaryExecutable(
            binary=REPEAT_DETECTOR_PATH[name],
            name=name)
        super(DetectorXStream, self).__init__(executable)

    def find_chartfile(self, searchdir):
        """Look for XSTREAM output xls file and return filepath"""
        files = os.listdir(searchdir)

        # look for the first file that matches XSTREAM_something_chart.xl
        pat_chartfilename = re.compile("^XSTREAM_.*_chart\.xls$")
        for f in files:
            if pat_chartfilename.match(f):
                return os.path.join(searchdir, f)
        else:
            return None

    def run_process(self, working_dir, infile):
        """ Run detector process on infile in working_dir and return all repeats found.

        Run detector process on infile in working_dir and return all repeats found.

        Args:
            working_dir (str): Working directory
            infile (str): Infile

        .. ToDo:: Complete docstrings.
        """

        wd = os.path.join(working_dir, DetectorXStream.name)

        # execute detector process
        retcode, stdoutfname, stderrfname = super(
            DetectorXStream, self).run_process(
            wd, *self.config.tokens(infile))

        if check_java_errors(stdoutfname, stderrfname,
                             log=LOG, procname=DetectorXStream.displayname):
            return []

        # Find output file, cry about it if not found

        #shutil.rmtree(os.path.join(EXECROOT, '..', 'spielwiese'))
        #shutil.copytree(working_dir, os.path.join(EXECROOT, '..', 'spielwiese'))

        chartfilename = self.find_chartfile(wd)
        if not chartfilename:
            LOG.error("No XSTREAM output chart file found in %s!", wd)
            return []

        # Process output file, return results
        with open(chartfilename, "r") as infile:
            return list(repeat_detection_io.xstream_get_repeats(infile))


def Detectors(lDetector=None, sequence_type=None):
    """ Define a global dictionary of all used detector functions.

    Define a global dictionary of all used detector functions.

    Args:
        lDetector (list of str): A list of repeat detection algorithm names.
        sequence_type (str): Either "AA" or "DNA".

    Raises:
        Exception: if at least one of the provided detectors in ``lDetector`` does not exist.

    """

    global DETECTORS

    if not sequence_type:
        sequence_type = CONFIG_GENERAL["sequence_type"]
    if not lDetector:
        lDetector = CONFIG_GENERAL["sequence"]["repeat_detection"][sequence_type]
    else:
        if isinstance(lDetector, str):
            lDetector = [lDetector]
        elif not isinstance(lDetector, list):
            raise TypeError(""" lDetector is not of type list. Please supply a list of TR detectors
            (e.g. lDetector = ['HHrepID']). If you use TR detectors defined in config.ini, make
            sure the TR detector list (AA or DNA) is correctly defined.""")
        if any(i not in FINDER_LIST.keys() for i in lDetector):
            raise Exception(
                "Unknown TR detector supplied (Supplied: {}. Known TR detectors: {})".format(
                    lDetector,
                    FINDER_LIST.keys()))

    DETECTORS = {FINDER_LIST[i].name: FINDER_LIST[i]() for i in lDetector}


def split_sequence(seq_records, working_dir):
    """ Split a FASTA file with multiple entries to several FASTA files with one entry

    Arguments:
    seq_records -- List of Sequence instances.
    working_dir -- Output directory for splitted file

    Returns:
    A list of tuples containing the Protein identifier and the file name

    .. TODO: Do not append i to outfiles, but a sequence ID
    """

    outfiles = []

    for i, seq in enumerate(seq_records):
        filename = "sequence_{0:03}.faa".format(i + 1)
        seq.write(file=os.path.join(working_dir, filename), file_format="fasta")
        outfiles.append((i, filename))

    return outfiles


class DetectorJob:

    def __init__(self, detector, infile, job_id, protein_id):
        self.detector = detector
        self.infile = infile
        self.job_id = job_id
        self.protein_id = protein_id


################ RUN A SET OF Tandem Repeat Detection algorithms (TRDs) ##

def run_detector(
        seq_records,
        detectors=None,
        sequence_type='AA',
        default=True,
        local_working_dir=None,
        num_threads=1):
    """ Run TRD on sequence_records and return the predicted repeats for each ``seq_records``
     and for each tandem repeat detector.

    Run TRD on sequence_records and return the predicted repeats for each ``seq_records``
    and for each tandem repeat detector. Example of return type::

        [
            # record 1
            {
            'T-REKS' : [ Repeat(), Repeat(), ...],
            'XSTREAM' : [ Repeat(), Repeat(), ...],
            ...
            },
            # record 2
            ...
        ]

    Args:
        seq_records (list of Sequence): A list of Sequence instances
        detectors (list of str): A list tandem repeat detector names
        sequence_type (str): Either "AA" or "DNA"
        default (bool): If True, default values for the detection algorithms are used.
        local_working_dir (str): Directory where data and results are stored. If provided,
        temporary files are not deleted/
        num_threads (int): Run ``num_threads`` detectors on parallel threads.

    Returns:
        list of dictionary: A list with a dictionary for each record in seq_records. The
        dictionary contains a list of repeats for each detector that was used.
    """

    # Create temporary working dir
    if local_working_dir:
        working_dir = local_working_dir
    else:
        working_dir = tempfile.mkdtemp()
        LOG.debug(
            "repeat_detection_run.run_detector: Created tempfile: %s",
            working_dir)
        if not os.path.isdir(working_dir):
            raise IOError("The specified directory \"" + working_dir +
                          "\" does not exist")

    # Initialise Detectors
    Detectors(detectors, sequence_type)

    # Adjust TRD parameters:
    if not default:
        if sequence_type == 'AA':
            DETECTORS['HHrepID'].config = set_hhrepid_config_open()
            DETECTORS['TRUST'].config = set_trust_config_open()
        else:
            DETECTORS['TRF'].config = set_trf_config_open()
            DETECTORS['PHOBOS'].config = set_phobos_config_open()
        DETECTORS['T-REKS'].config = set_treks_config_open()
        DETECTORS['XSTREAM'].config = set_xstream_config_open()

    if sequence_type == 'DNA':
        DETECTORS['T-REKS'].config = set_treks_config_DNA()

    infiles = split_sequence(seq_records, working_dir)

    # list of dictionaries for each protein containing detectorname : results
    # entries
    results = [
        {fname: [] for fname, detector in DETECTORS.items()}
        for i in range(len(infiles))
    ]

    # Create a list for our jobs
    joblist = [
        DetectorJob(
            detector=detector,
            infile=os.path.join(working_dir, infile[1]),
            job_id=job_id,
            protein_id=infile[0]
        )
        for fname, detector in DETECTORS.items()
        for job_id, infile in enumerate(infiles)
    ]

    LOG.info("Processing %d input files in %d jobs.",
             len(infiles), len(joblist))

    # put joblist into job queue, thus starting actual work
    for job in joblist:
        try:
            wd = os.path.join(working_dir, "{0:03}".format(job.job_id + 1))

            LOG.debug("Launching detector \"%s\" in directory %s",
                      job.detector.name, os.path.join(wd, job.detector.name))

            result = job.detector.run_process(wd, job.infile)

            # lock the mutex for results, append result
            # with result_lock:
            results[job.job_id][job.detector.name].extend(result)

            LOG.debug("Detector \"%s\" returned from job %d",
                      job.detector.name, job.job_id)
        except:
            LOG.exception(
                "Exception occured in worker while processing %s with %s",
                job.infile, job.detector.name)

    LOG.info("All jobs returned.")

    # delete temporary directory
    if not local_working_dir:
        try:
            shutil.rmtree(working_dir)
        # I guess the error type is known, and you could be more precise :)
        except:
            LOG.error("Unexpected error: {0}".format(sys.exc_info()[0]))
            raise

    return results


######## SET OPEN CONFIGS #########

def set_hhrepid_config_open():
    # construct open configuration for HHrepid
    config = DetectorHHrepID.Configuration()
    config.boolopts["-nofilt"] = True
    # <float> max p-value of suboptimal alignments in all search rounds but the last one (def=0.1)
    config.valopts["-P"] = 0.6
    config.valopts["-T"] = 0.2  # <float> max total repeat p-value (def=0.001)
    # [0,inf[  minimal length of repeats to be identified (def=7)
    config.valopts["-lmin"] = 0
    LOG.debug("%s config tokens: %s", DETECTORS['hhrepid'].displayname,
              ", ".join(DETECTORS['hhrepid'].config.tokens()))
    return config


def set_phobos_config_open():
    # construct open configuration for Phobos
    config = DetectorPhobos.Configuration()
    config.valopts["--maxUnitLen"] = 100
    config.valopts["--mismatchScore"] = -2
    config.valopts["--indelScore"] = -2
    return config

# Incomplete


def set_treks_config_DNA():
    # construct DNA configuration for T-Reks
    config = DetectorTREKS.Configuration()
    config.valopts["-type"] = 1
    LOG.debug("%s config tokens: %s", DETECTORS['t-reks'].displayname,
              ", ".join(DETECTORS['t-reks'].config.tokens()))
    return config


def set_treks_config_open():
    # construct open configuration for T-Reks
    config = DetectorTREKS.Configuration()
    config.valopts["-similarity"] = 0.2
    LOG.debug("%s config tokens: %s", DETECTORS['t-reks'].displayname,
              ", ".join(DETECTORS['t-reks'].config.tokens()))
    return config


def set_trf_config_open():
    # construct open configuration for TRF
    config = DetectorTRF.Configuration()
    config.valopts["Minscore"] = 20
    LOG.debug("%s config tokens: %s", DETECTORS['trf'].displayname,
              ", ".join(DETECTORS['trf'].config.tokens()))
    return config


def set_trust_config_open():
    # construct open configuration for Trust
    config = DetectorTrust.Configuration()
    config.boolopts['-force'] = True
    LOG.debug("%s config tokens: %s", DETECTORS['trust'].displayname,
              ", ".join(DETECTORS['trust'].config.tokens()))
    return config


def set_xstream_config_open():
    # construct open configuration for XStream
    config = DetectorXStream.Configuration()
    #config.boolopts[optname] = True
    config.valopts["-I"] = 0.2
    config.valopts["-i"] = 0.1
    config.valopts["-g"] = 100
    LOG.debug("%s config tokens: %s", DETECTORS['xstream'].displayname,
              ", ".join(DETECTORS['xstream'].config.tokens()))
    return config


######################## HARDCODED OVERVIEW DICTIONARIES #################

FINDER_LIST = {"HHrepID": DetectorHHrepID,
               "Phobos": DetectorPhobos,
               "TRED": DetectorTRED,
               "T-REKS": DetectorTREKS,
               "TRF": DetectorTRF,
               "TRUST": DetectorTrust,
               "XSTREAM": DetectorXStream
               }



#! /usr/bin/env python

import configobj
import os
import os.path
import re
import shlex
import shutil
import sqlalchemy as sqla
import sys

# Interface to Gc3libs

import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, _Script
from gc3libs.workflow import RetryableTask, SequentialTaskCollection, ParallelTaskCollection
from gc3libs.persistence.accessors import GetValue
from gc3libs.quantity import kB, MB, GB
import gc3libs.debug
import gc3libs.utils


######################## Basic Applications/Tasks Templates ##############

class MyApplication(Application):

    """
    Basic template method pattern  `Application`:class: Initialise Application generically,
    and check for successful running generically.
    """
    #@gc3libs.debug.trace

    def __init__(self, name, **kwargs):

        gc3libs.log.info("Initialising {}".format(self.__class__.__name__))
        config = kwargs["config"]
        self.c = config[name].copy()
        self.jobname = self.__class__.__name__  # Used by GC3PIE

        kwargs['output_dir'] = self.c['logdir']

        # Replace every "%X" in the config with the current value for X, e.g.
        # "3".
        if "param" in kwargs:
            for iC in self.c.keys():
                for param_name, param_value in kwargs['param'].items():
                    self.c[iC] = self.c[iC].replace(param_name, param_value)

            for param_name, param_value in kwargs['param'].items():
                if param_name == '$N':
                    self.N = param_value
                # Adapt output_dir to particular file
                kwargs['output_dir'] = kwargs['output_dir'].replace(
                    param_name,
                    param_value)

        if "required_memory" in self.c:
            kwargs['requested_memory'] = int(self.c['required_memory']) * MB
        else:
            kwargs['requested_memory'] = int(config['required_memory']) * MB

        gc3libs.Application.__init__(
            self,
            arguments=shlex.split(
                self.c['script']) +
            ["-i"] +
            shlex.split(
                self.c['input']) +
            ["-o"] +
            shlex.split(
                self.c['output']) +
            shlex.split(
                self.c['extra']),
            inputs=[],
            outputs=[],
            join=True,
            stdout=self.c['stdout'],
            stderr=self.c['stderr'],
            **kwargs)

    def terminated(self):
        gc3libs.log.info(
            "Testing whether {} is done".format(
                self.__class__.__name__))

        # If the application has terminated o.k. (self.execution.returncode == 0),
        #   Check wether all resultfiles are o.k. If yes: Good, If no: FREEZE.
        # If not: Save error and FREEZE.

        if self.execution.returncode == 0:
            gc3libs.log.info(
                "{1} claims to be successful: self.execution.returncode: {0}".format(
                    self.execution.returncode,
                    self.__class__.__name__))
            # Check if result file exists (Later: and complies to some rules).
            if self.is_valid(self.c['output']):  # IMPLEMENT IN CLASS!
                gc3libs.log.info(
                    "{} has run successfully to completion.".format(
                        self.__class__.__name__))
                # Now, clean up
                # Delete all log files if everything ran smoothly.
                shutil.rmtree(self.output_dir)
            else:
                gc3libs.log.info(
                    "%s has not produced a valid outputfile.",
                    self.__class__.__name__)
                # Set self.execution.exitcode to a non-zero integer <256 (to
                # indicate an error)
                self.execution.exitcode = 42

        else:
            gc3libs.log.info(
                "{1} is not successful: self.execution.returncode: {0}".format(
                    self.execution.returncode,
                    self.__class__.__name__))
            # Check if there is stderr.
            if not os.path.isfile(os.path.join(self.output_dir, self.stderr)):
                self.error_tag = ""
            else:
                # Create a tag from the last line in stderr.
                with open(os.path.join(self.output_dir, self.stderr), "r") as fh:
                    for line in fh:
                        pass
                    self.error_tag = line
            self.execution.exitcode = 88


class StopOnError(object):

    """
    Mix-in class to make a `SequentialTaskCollection`:class: turn to STOPPED
    state as soon as one of the tasks fail.
    """

    def next(self, done):
        if done == len(self.tasks) - 1:
            self.execution.returncode = self.tasks[done].execution.returncode
            return Run.State.TERMINATED
        else:
            rc = self.tasks[done].execution.exitcode
            if rc != 0:
                return Run.State.STOPPED
            else:
                return Run.State.RUNNING


################################# Applications/Tasks #####################

class FilePreparation(MyApplication):

    def is_valid(self, output):

        #        if not os.path.isfile(self.c['output']):
        #        gc3libs.log.info("{} has not produced an output file.".format(self.__class__.__name__))
        #        self.execution.returncode = 1
        #        # Set self.execution.exitcode to a non-zero integer <256 (to indicate an error)
        #        self.execution.exitcode = 42

        return True


class AnnotateTandemRepeats(MyApplication):

    def is_valid(self, output):
        return True


class SerializeAnnotations(MyApplication):

    def is_valid(self, output):
        return True


############################# Main Session Creator (?) ###################


class TandemRepeatAnnotationWorkflow(SessionBasedScript):

    """
    TandemRepeatAnnotationWorkflow
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version='0.0.1',
        )

    def setup_options(self):
        self.add_param("-conf", "--config_file", type=str, required=True,
                       help="Path to the parameter file")

    def _make_session(self, session_uri, store_url):
        return gc3libs.session.Session(
            session_uri,
            store_url,
            extra_fields={
                # NB: enlarge window to at least 150 columns to read this table
                # properly!
                # Former: Class                                           , #
                # task class
                sqla.Column('jobname', sqla.TEXT()): (lambda obj: obj.__class__.__name__),
                # .ONLY(CodemlApplication), # program executable
                sqla.Column('executable', sqla.TEXT()): GetValue(default=None) .arguments[0],
                # .ONLY(CodemlApplication), # fullpath to codeml output directory
                sqla.Column('output_path', sqla.TEXT()): GetValue(default=None) .output_dir,
                sqla.Column('cluster', sqla.TEXT()): GetValue(default=None) .execution.resource_name,  # .ONLY(CodemlApplication), # cluster/compute element
                # sqla.Column('worker',             sqla.TEXT())    : GetValue(default=None) .hostname                          ,#.ONLY(CodemlApplication), # hostname of the worker node
                # sqla.Column('cpu',                sqla.TEXT())    : GetValue(default=None) .cpuinfo                           ,#.ONLY(CodemlApplication), # CPU model of the worker node
                # sqla.Column('requested_walltime', sqla.INTEGER()) : _get_requested_walltime_or_none                           , # requested walltime, in hours
                sqla.Column('requested_cores', sqla.INTEGER()): GetValue(default=None) .requested_cores,  # .ONLY(CodemlApplication), # num of cores requested
                sqla.Column('duration_s', sqla.INTEGER()): lambda app: app.execution.duration.amount(unit=app.execution.duration.second),  # .ONLY(CodemlApplication), # used walltime
                sqla.Column('max_used_memory_mb', sqla.INTEGER()): lambda app: app.execution.max_used_memory.amount(unit=app.execution.max_used_memory.MB),  # .ONLY(CodemlApplication), # used max_used_memory # might need to be saved as TEXT
                sqla.Column('lrms_jobid', sqla.TEXT()): GetValue(default=None) .execution.lrms_jobid,  # .ONLY(CodemlApplication), # arc job ID
                # sqla.Column('original_exitcode',  sqla.INTEGER()) : GetValue(default=None) .execution.original_exitcode       ,#.ONLY(CodemlApplication), # original exitcode
                sqla.Column('used_cpu_time_s', sqla.INTEGER()): lambda app: app.execution.used_cpu_time.amount(unit=app.execution.used_cpu_time.second),  # .ONLY(CodemlApplication), # used walltime
                # returncode = exitcode*256 + signal
                sqla.Column('returncode', sqla.INTEGER()): GetValue(default=None) .execution.returncode,  # .ONLY(CodemlApplication), # returncode attr
                # sqla.Column('queue',              sqla.TEXT())    : GetValue(default=None) .execution.queue                   ,#.ONLY(CodemlApplication), # exec queue _name_
                sqla.Column('time_submitted', sqla.FLOAT()): GetValue(default=None) .execution.timestamp['SUBMITTED'],  # .ONLY(CodemlApplication), # client-side submission (float) time
                sqla.Column('time_terminated', sqla.FLOAT()): GetValue(default=None) .execution.timestamp['TERMINATED'],  # .ONLY(CodemlApplication), # client-side termination (float) time
                sqla.Column('time_stopped', sqla.FLOAT()): GetValue(default=None) .execution.timestamp['STOPPED'],  # .ONLY(CodemlApplication), # client-side stop (float) time
                sqla.Column('error_tag', sqla.TEXT()): GetValue(default=None) .error_tag,
                sqla.Column('N', sqla.TEXT()): GetValue(default=None).N.ONLY((tandem_repeat_annotation_workflow.AnnotateTandemRepeats)),
            })

    def parse_args(self):
        self.config_file = self.params.config_file
        gc3libs.log.info(
            "TestWorkflow config_file: {}".format(
                self.config_file))

    def new_tasks(self, kwargs):

        #name = "myTestWorkflow"
        gc3libs.log.info("Calling TestWorkflow.next_tasks()")

        self.config = configobj.ConfigObj(self.config_file, stringify=True)

        yield tandem_repeat_annotation_workflow.MainSequentialFlow(config=self.config, **kwargs)


######################## Support Classes / Workflow elements #############


class MainSequentialFlow(SequentialTaskCollection):

    def __init__(self, **kwargs):

        config = kwargs["config"]
        self.kwargs = kwargs
        gc3libs.log.info(
            "\t Calling MainSequentialFlow.__init({})".format("<No parameters>"))

        self.initial_tasks = []
        if config["file_preparation"]["activated"] == 'True':
            self.initial_tasks = [
                FilePreparation(
                    name="file_preparation",
                    **kwargs)]
        else:
            self.initial_tasks = [SequencewiseParallelFlow(**kwargs)]

        SequentialTaskCollection.__init__(self, self.initial_tasks, **kwargs)

    def next(self, done):
        last_task = self.tasks[done]
        rc = last_task.execution.exitcode
        gc3libs.log.info(
            "\t MainSequentialFlow: Find next task. last_task.execution.exitcode: [%s]" %
            rc)
        # Stop the execution if last application failed.
        if rc != 0:
            gc3libs.log.info("\t rc is not 0 [%s]" % type(rc))
            self.execution.rc = rc
            return Run.State.STOPPED

        # Shall we continue with the workflow?
        if last_task != self.tasks[-1]:
            gc3libs.log.info("\t MainSequentialFlow: Do not add anything")
            # We still have tasks to run in self.tasks, let's consume them
            # before adding new tasks.
            return Run.State.RUNNING
        elif isinstance(last_task, DataPreparationParallelFlow) or isinstance(last_task, FilePreparation):
            gc3libs.log.info(
                "\t MainSequentialFlow: Add SequencewiseParallelFlow")
            # SequencewiseParallelFlow task needs to be initialized using the
            # output from previous steps.
            self.add(SequencewiseParallelFlow(**self.kwargs))
            return Run.State.RUNNING
        else:
            gc3libs.log.info("\t MainSequentialFlow: Terminated")
            # Workflow is terminated.
            self.execution.rc = rc
            return Run.State.TERMINATED

    def terminated(self):
        self.execution.returncode = 0
        gc3libs.log.info(
            "\t MainSequentialFlow.terminated [%s]" %
            self.execution.returncode)


class DataPreparationParallelFlow(ParallelTaskCollection):

    def __init__(self, **kwargs):

        self.kwargs = kwargs
        gc3libs.log.info(
            "\t\tCalling DataPreparationParallelFlow.__init({})".format(
                self.kwargs))

        # Needs to be coded


class SequencewiseParallelFlow(ParallelTaskCollection):

    def __init__(self, **kwargs):

        config = kwargs["config"]
        self.c = config["sequencewise_parallel_flow"]

        # TODO: Find all files in dir and create self.lSeq! Warning! Should be done once the
        # Tasks before are finished.
        #self.lSeq = [re.findall(self.c['retag'], i)[0] for i in os.listdir(self.c['input'])]
        self.lSeq = [
            i for i in os.listdir(
                self.c['input']) if not i.endswith("fai")]
        self.kwargs = kwargs

        gc3libs.log.info(
            "\t\tCalling SequencewiseParallelFlow.__init({})".format(
                self.kwargs))

        self.tasks = [
            AnnotateTandemRepeats(
                name="annotate_tandem_repeats",
                param={
                    "$N": iSeq},
                **kwargs) for iSeq in self.lSeq]

        ParallelTaskCollection.__init__(self, self.tasks, **kwargs)

    def terminated(self):
        self.execution.returncode = 0
        gc3libs.log.info("\t\tSequencewiseParallelFlow.terminated")


def _get_requested_walltime_or_none(job):
    if isinstance(job, gc3libs.application.codeml.CodemlApplication):
        return job.requested_walltime.amount(hours)
    else:
        return None


# run script
if __name__ == '__main__':
    import tandem_repeat_annotation_workflow
    TandemRepeatAnnotationWorkflow().run()



# (C) 2014-2015 Elke Schaper

"""
    :synopsis: Configuration Singleton class, ensuring that the configuration
        is created in exactly one instance, even though it is called from
        multiple modules in arbitrary orders.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import configobj
from tral.paths import config_file

P_CONFIG = config_file("config.ini")
P_SPEC = config_file("spec.ini")


class Singleton:

    """
    A non-thread-safe helper class to ease implementing singletons.
    This should be used as a decorator -- not a metaclass -- to the
    class that should be a singleton.

    The decorated class can define one `__init__` function that
    takes only the `self` argument. Other than that, there are
    no restrictions that apply to the decorated class.

    To get the singleton instance, use the `Instance` method. Trying
    to use `__call__` will result in a `TypeError` being raised.

    Limitations: The decorated class cannot be inherited from.

    The credit goes to http://stackoverflow.com/users/627005/paul-manta
    http://stackoverflow.com/questions/42558/python-and-the-singleton-pattern

    """

    def __init__(self, decorated):
        self._decorated = decorated

    def instance(self):
        """
        Returns the singleton instance. Upon its first call, it creates a
        new instance of the decorated class and calls its `__init__` method.
        On all subsequent calls, the already created instance is returned.

        """
        try:
            return self._instance
        except AttributeError:
            self._instance = self._decorated()
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `Instance()`.')

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated)


@Singleton
class Configuration:
    """
    A singleton configuration object containing a `ConfigObj` instance.

    Attributes:
        config (configobj.ConfigObj): The `ConfigObj` instance.
    """
    def __init__(self):
        self.config = configobj.ConfigObj(
            P_CONFIG,
            configspec=P_SPEC,
            stringify=True)



import pytest
import sys

@pytest.mark.no_external_software_required
def test_imports():
    import argparse
    import Bio
    import configobj
    import docutils
    import numpy
    import pytest
    import scipy
    import setuptools
    import sphinx
    import scipy.linalg,scipy.special,scipy.stats

@pytest.mark.no_external_software_required
def test_python_version():

    # Check if Python3 is used
    assert sys.version_info[0] == 3




import collections
import os
import pytest

from tral.repeat_list import repeat_list as rl
from tral.repeat import repeat
from tral.repeat_list import repeat_list_io as rl_io
from tral.sequence import sequence

TEST_REPEATS = [["AA","AA"],["AAA","AAA"]]
TEST_SEQUENCE = "MAAAAKAAAAAAL"
TEST_SCORE = "phylo_gap01"
TEST_SCORE_VALUE_LIST = [0.0, 0.5]
TEST_BEGIN_LIST = [6,10]
# The resulting string should contain the following data, however perhaps in a different order:
TEST_TSV = "msa_original\tbegin\tn_effective\tl_effective\tsequence_length\tpvalue\nAA,AA\t2\t2.0\t2\t4\tNone\nAAA,AAA\t7\t2.0\t3\t6\tNone"


@pytest.mark.no_external_software_required
def test_serialize_repeat_list_tsv():

    test_repeats = [repeat.Repeat(msa = i) for i in TEST_REPEATS]
    test_seq = sequence.Sequence(TEST_SEQUENCE)
    for i in test_repeats:
        test_seq.repeat_in_sequence(i)
    test_repeat_list = rl.RepeatList(repeats = test_repeats)

    tsv = rl_io.serialize_repeat_list_tsv(test_repeat_list)

    assert type(tsv) == str



# (C) 2011, Alexander Korsunsky
# (C) 2011-2015 Elke Schaper

"""
    :synopsis: Parsing repeat detection algorithm output

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import logging
import re
import csv

LOG = logging.getLogger(__name__)


class RepeatRegion:

    def __init__(self, protein_id="", begin=None, msa=None):
        self.protein_id = protein_id
        self.begin = begin
        if msa is None:
            msa = []
        self.msa = msa


def tred_get_repeats(infile):
    """ Read repeats from a TRED standard output (stdout) file stream successively.

    Read repeats from a TRED standard output (stdout) file stream successively.
    Postcondition: infile points to EOF.

    Layout of TRED output file::

         Start: start End: \d+ Length: \d+

        ( \d repeat_unit \d
        ( alignment_indicator )?)*

    Args:
        infile (file stream): File stream of output1 from
            tred1 myDNA.faa intermediate_output
            tred2 myDNA.faa intermediate_output output1 output2

    Returns:
         (Repeat): A generator function is returned that, when called in a loop,
         yields one repeat per iteration.

    .. todo:: Layout TRED output syntax.
    """

    # pattern for start index
    pat_start = re.compile(r"Start: (\d+)")
    pat_repeat_unit = re.compile(r"\d+\s+([ACGT-]+)")
    pat_alignment_indicator = re.compile(r"\s+([ACGT-]+)")

    # Our possible parser states:
    #
    # 1: searching for start
    # 2: searching for first repeat unit
    # 3: searching for all further repeat units

    identifier = ''
    state = 1
    repeat_units = []
    for i, line in enumerate(infile):
        LOG.debug("Line %d: %s", i, line[0:-1])
        if 1 == state:
            region = RepeatRegion(protein_id=identifier, msa=[])
            match = pat_start.match(line)
            if match:
                LOG.debug(" * (1->2) Found start")
                LOG.debug("Start: %s", match.group(1))
                region.begin = int(match.group(1))
                state = 2

        if 2 == state:
            match = pat_repeat_unit.match(line)
            if match:
                LOG.debug(" * (2->3) Found first repeat_unit")
                repeat_units.append((match.group(1), True))
                state = 3
                continue

        if 3 == state:
            match = pat_repeat_unit.match(line)
            if match:
                LOG.debug(" * (3->3) Found another repeat_unit")
                repeat_units.append((match.group(1), True))
            else:
                match = pat_alignment_indicator.match(line)
                if match:
                    LOG.debug(" * (3->3) Found an alignment_indicator unit")
                    repeat_units.append((match.group(1), False))
                else:
                    LOG.debug(" * (3->1) Found end of repeat (yielding)")
                    state = 1
                    region.msa = tred_msa_from_pairwise(repeat_units)
                    yield region


def tred_msa_from_pairwise(repeat_units):
    """ Construct a MSA from pairwise alignments.

    Construct a MSA from pairwise alignments. At the moment, gaps following the repeat are
    not added. However, these gaps are added automatically when a ``Repeat`` instance is
    created.

    Args:
        repeat_units (list of str): Read in from TRED output files

    Returns:
         (list of str)

    .. todo:: Is the Args format mentioned correctly?
    """

    pat_gap = re.compile(r"(-*)")
    result = []
    index = 0
    for iR in range(len(repeat_units)):
        ru = repeat_units[iR]

        # The next repeat unit
        if ru[1] == True:
            result.append('-' * index + ru[0])
            # How many gaps in the beginning of this repeat unit?
            index += len(pat_gap.match(ru[0]).group())

        # The next alignment indicator
        else:
            for iL in range(len(ru[0])):
                if ru[0][iL] == '-':
                    # enter a gap between
                    # the index + iL and index + iL + 1 character
                    # in each repeat unit in result so far:
                    result = [iRU[:index + iL] + '-' + iRU[index + iL:]
                              for iRU in result]

    return result


def treks_get_repeats(infile):
    """ Read repeats from a T-REKS standard output (stdout) file stream successively.

    Read repeats from a T-REKS standard output (stdout) file stream successively.
    Postcondition: infile points to EOF.

    Layout of T-REKS output file::

        protein ::=
            ">" identifier
            repeat*
        #
        repeat ::=
            repeat_header
            sequence*
            "*"+
        #
        repeat_header ::= "Length:" integer "residues - nb:" integer  "from"  integer "to" integer "- Psim:"float "region Length:"integer


    Args:
        infile (file stream): File stream to the file of the standard output of T-Reks

    Returns:
         (Repeat): A generator function is returned that, when called in a loop,
         yields one repeat per iteration.

    .. todo:: Layout T-REKS output syntax.
    """

    # pattern for protein identifier
    pat_identifier = re.compile(r">(\S+)")

    # pattern for repeat properties
    pat_repeat_header = re.compile(
        r"Length: \d+ residues - nb: (\d+)  from  (\d+) to (\d+) - Psim:([\d\.]+) region Length:(\d+)")

    pat_repeat_end = re.compile(r"\*+")

    # pattern for repeat sequence
    # FIXME stuff that occurs here is not just A-Z but a set of possible amino
    # acid symbols
    pat_sequence = re.compile(r"([A-Z-]+)")

    # Our possible parser states:
    #
    # 1: state between repeats
    #   entry: reset repeat
    #   expect repeat_header(goto 2) OR identifier(store identifier, goto 1)
    # 2: state for multiple sequence alignment line
    # expect sequence(goto 3, append sequence) OR repeat_end(return repeat,
    # goto 1)

    state = 1
    identifier = ""

    for i, line in enumerate(infile):
        LOG.debug("Line %d: %s", i, line[0:-1])

        if 1 == state:
            region = RepeatRegion(protein_id=identifier, msa=[])
            LOG.debug("msa: %s", "\n".join(region.msa))
            match = pat_repeat_header.match(line)
            if match:
                LOG.debug(" * (1->2) Found properties")
                region.begin = int(match.group(2))
                state = 2

            match = pat_identifier.match(line)
            if match:
                LOG.debug(" * (1->1) Found identifier")
                identifier = match.group(1)
                state = 1

        elif 2 == state:
            match = pat_sequence.match(line)
            if match:
                LOG.debug(" * (2->2) Found MSA line (appending)")
                region.msa.append(match.group(1))
                state = 2
            match = pat_repeat_end.match(line)
            if match:
                LOG.debug(" * (2->1) Found end of repeat (yielding)")
                state = 1
                yield region

        else:
            raise AssertionError("Huh? Unknown parser state " +
                                 str(state))


def xstream_get_repeats(infile):
    """ Read repeats from a XSTREAM output xls chart

    Read repeats from a XSTREAM output xls chart

    Postcondition: infile points to EOF.

    Args:
        infile (file stream): File stream to read XSTREAM output from a xls chart.

    Returns:
         (Repeat): A generator function is returned that, when called in a loop,
         yields one repeat per iteration.
    """

    # The infile is luckily enough in csv format:
    reader = csv.reader(infile, dialect='excel-tab')

    # read first row with the fieldnames
    header = next(reader)

    LOG.debug("Header has %d fields.", len(header))

    if len(header) != 8 and len(header) != 10:
        raise ValueError(
            "XStream output file seems to be malformed. "
            "Make sure to open this function with the generated xls chart.")

    # when XSTREAM is fed files with multiple sequences, first field is "identifier",
    # otherwise it is "seq length"
    if header[0] == "identifier":
        field_offset = 2
    else:
        field_offset = 0

    for row in reader:
        region = RepeatRegion()
        if field_offset:
            region.protein_id = row[0]

        region.begin = int(row[0 + field_offset])
        region.msa = row[4 + field_offset].split()
        LOG.debug("Found repeat, yielding.")
        if len(region.msa) >= 2:
            yield region


def trust_fill_repeats(msa, begin, sequence, maximal_gap_length=20):
    ''' return a trust msa that has no longer indels than maximal_gap_length,
    that contains the indel characters even when not part of the trust output file.
    Background trust returns tandem repeats, but also distant repeats. '''
    gapless_msa = [repeat_unit.replace('-', '').upper() for repeat_unit in msa]
    sequence = sequence.upper()

    # Find the start and end positions of the predicted repeat units
    position = [(begin - 1, begin + len(gapless_msa[0]) - 2)]
    for repeat_unit in gapless_msa[1:]:
        find_index = sequence[position[-1][1] + 1:].find(repeat_unit)
        # Repeat unit could not be found in sequence -> Discard the repeat
        if find_index == -1:
            return None, None
        repeat_unit_begin = find_index + position[-1][1] + 1
        position.append(
            (repeat_unit_begin, repeat_unit_begin + len(repeat_unit) - 1))

    # Derive the start and end positions of the gaps
    gap_position = [(i[1] + 1, j[0] - 1)
                    for i, j in zip(position[:-1], position[1:])]
    gaps = [i[1] - i[0] + 1 for i in gap_position]

    # Filter out repeat units that are further apart than maximal_gap_length
    gap_valid = ''.join(
        ['1' if i_gap <= maximal_gap_length else '0' for i_gap in gaps])

    count_valid_pairs = [len(m.group())
                         for m in re.finditer(re.compile('1+'), gap_valid)]
    # All repeat units are further apart than maximal_gap_length? -> Discard
    # the repeat
    if len(count_valid_pairs) == 0:
        return None, None

    # Choose the sequence of pairs closer then maximal_gap_length that is
    # longest
    valid_index = gap_valid.find('1' * max(count_valid_pairs))

    # Shorten the predicted msa accordingly.
    msa = msa[valid_index:valid_index + max(count_valid_pairs) + 1]
    gaps = gaps[valid_index:valid_index + max(count_valid_pairs) + 1]
    gap_position = gap_position[
        valid_index:valid_index +
        max(count_valid_pairs) +
        1]
    position = position[valid_index:valid_index + max(count_valid_pairs) + 1]

    # Add missing sequence to the repeat units
    repeat_unit_length = len(msa[0])
    gap_count_before = 0
    for i, i_gap in enumerate(gaps):
        gap_count_after = gap_count_before + i_gap
        msa[i] += '-' * (gap_count_before) + sequence[gap_position[i][0]:gap_position[i][1] + 1] + '-' * (sum(gaps) - gap_count_after)
        gap_count_before = gap_count_after
    msa[-1] += '-' * sum(gaps)

    return msa, position[0][0] + 1


def trust_get_repeats(infile):
    """ Read repeats from a TRUST standard output (stdout) file stream successively.

    Read repeats from a TRUST standard output (stdout) file stream successively.
    Postcondition: infile points to EOF.

    Layout of TRUST standard output::

        protein ::=
            ">" identifier
            (repeat_types)*
            "//"
        #
        repeat_types ::=
            "REPEAT_TYPE" integer
            "REPEAT_LENGTH" integer
            (repeat_info)*
            (">Repeat " integer
            sequence)*
        #
        repeat_info ::=
            integer integer [integer] [integer]

    Args:
        infile (file stream): File stream from TRUST standard output.

    Returns:
         (Repeat): A generator function is returned that, when called in a loop,
         yields one repeat per iteration.

    .. todo:: Layout TRUST output syntax.
    """

    pat_identifier = re.compile(r">(\S+)")
    pat_repeat_type = re.compile(r"REPEAT_TYPE \d+")
    pat_repeat_length = re.compile(r"REPEAT_LENGTH \d+")
    pat_repeat_info = re.compile(r"(\d+) (\d+).*$")
    pat_repeat_header = re.compile(r">Repeat \d+")
    # FIXME find proper character set here
    pat_repeat_sequence = re.compile("([A-Za-z-]+)")
    pat_protein_end = re.compile("//")

    # Our possible parser states:
    #
    # 0: initial state
    #   expect: identifier(store identifier, goto 1)
    # 1: before the beginning of a repeat
    #   expect: "REPEAT_TYPE"(goto 2) or "//"(goto 0)
    # 2: state after having found "REPEAT_TYPE"
    #   entry: reset state of return value (repeat)
    #   expect: "REPEAT_LENGTH"(goto 3)
    # 3: first line of repeat info
    #   expect: repeat_info(store begin, goto 4) or ">Repeat"(goto 5)
    # 4: continuing to read repeat info
    #   expect: repeat_info(goto 4) or ">Repeat"(goto 5)
    # 5: part of MSA sequence
    #   expect: sequence(goto 6)
    # 6: After first sequence
    # expect: ">Repeat"(goto 5) or "REPEAT_TYPE"(return repeat, goto 2) or
    # "//"(return repeat, goto 0)

    state = 0
    region = RepeatRegion()
    identifier = ""

    def strip_comments(line):
        pat_comment = re.compile(r"\s*#.*$")
        return pat_comment.sub("", line)

    for i, line in enumerate(infile):
        line = strip_comments(line)
        if not line or line == '\n':
            continue

        LOG.debug("Line %d: %s", i, line[0:-1])

        if 0 == state:
            match = pat_identifier.match(line)
            if match:
                LOG.debug(
                    " *(0->1) Found identifier (storing \"%s\")",
                    match.group(1))

                identifier = match.group(1)
                state = 1
                continue

        elif 1 == state:
            match = pat_repeat_type.match(line)
            if match:
                LOG.debug(" *(1->2) Found REPEAT_TYPE")
                state = 2
                continue

            match = pat_protein_end.match(line)
            if match:
                LOG.debug(" *(1->0) Found protein end")
                state = 0
                continue

        elif 2 == state:
            # Entry Action: clear msa cache
            region = RepeatRegion(protein_id=identifier)

            match = pat_repeat_length.match(line)
            if match:
                LOG.debug(" *(2->3) Found REPEAT_LENGTH")
                state = 3
                continue

        elif 3 == state:
            match = pat_repeat_info.match(line)
            if match:
                LOG.debug(
                    " *(3->4) Found repeat_info (storing begin: \"%s\")",
                    match.group(1)
                )

                region.begin = int(match.group(1))
                state = 4
                continue

            match = pat_repeat_header.match(line)
            if match:
                LOG.debug(" *(3->5) Found repeat_header")
                state = 5
                continue

        elif 4 == state:
            match = pat_repeat_info.match(line)
            if match:
                LOG.debug(" *(4->4) Found repeat_info")
                state = 4
                continue

            match = pat_repeat_header.match(line)
            if match:
                LOG.debug(" *(4->5) Found repeat_header")
                state = 5
                continue

        elif 5 == state:
            match = pat_repeat_sequence.match(line)
            if match:
                LOG.debug(
                    " *(5->6) Found sequence (storing \"%s\")",
                    match.group(1))

                region.msa.append(match.group(1))

                state = 6
                continue

        elif 6 == state:
            match = pat_repeat_header.match(line)
            if match:
                LOG.debug(" *(6->5) Found repeat_header")
                state = 5
                continue
            match = pat_repeat_type.match(line)
            if match:
                LOG.debug(" *(6->2) Found REPEAT_TYPE, yielding.")

                state = 2
                if len(region.msa) >= 2:
                    yield region
                continue
            match = pat_protein_end.match(line)
            if match:
                LOG.debug(" *(6->0) Found protein end, yielding.")
                state = 0
                if len(region.msa) >= 2:
                    yield region
                continue

        else:
            raise AssertionError("Huh? Unknown parser state " + str(state))

################################## TRF - Benson ##########################


def trf_get_repeats(infile):
    """ Read repeats from a TRF txt.html file stream file stream successively.

    Read repeats from a TRF txt.html file stream file stream successively.
    Postcondition: infile points to EOF.

    TRF output file syntax::

        Sequence: ``identifier``
             Indices: ``begin``--``end``
             \d [a-zA-Z]+
        #
             begin (repeat)*
             1  (consensus)*
        #
          (( \d (repeat)*
             \d  (consensus)*
          )?
             \d (repeat)*
             1  (consensus)*
          )+
             \d [a-zA-Z]+
        #
            ``Statistics``

    Args:
        infile (file stream): File stream from TRF output txt.html.
    (generated via e.g. ./trf404.linux64.exe FASTAFILE 2 7 7 80 10 50 500 -d > /dev/null
    If the -h flag is set, no .txt.html output is produced)

    Returns:
         (Repeat): A generator function is returned that, when called in a loop,
         yields one repeat per iteration.

    .. todo:: Layout TRF output syntax.
    .. todo:: Does not search for the sequence identifier at current!
    """

    # find the name of the sequence ## CURRENTLY NOT IMPLEMENTED
    #pat_identifier = re.compile("Sequence: (\S+)")

    # find coordinates of the repeat region in the protein
    pat_coordinates = re.compile("Indices: (\d+)--(\d+)")

    # find a part of a repeat unit and its coordinate
    pattern_seq = re.compile("\s+(\d+) ([ACGT\- ]+)")

    # find the final tag 'Statistics'
    pat_statistics = re.compile("Statistics")

    # Our possible parser states:
    #
    # state 0: searching for identifier -> 1  # not necessary when sequence identifier is known
    # state 1: searching for repeat region coordinates -> 2
    # state 2: searching for beginning of MSA & save sequence to tmpMSA-> 4
    # state 3: new repeat unit: save sequence to tmpMSA -> 6
    # state 4: new repeat unit: save sequence to tmp_consensus -> 5
    # state 5:
    #           if sequence: save sequence to tmpMSA -> 6
    #           if end: save tmpMSA to preMSA; save tmp_consensus to consensus; Yield repeat region -> 1
    # state 6: check: new repeat unit? save tmp_consensus
    #  1: use last tmpMSA entry for new tmpMSA;
    #     save tmpMSA to preMSA;
    #     save tmp_consensus to consensus;
    #     save sequence to new tmp_consensus -> 3
    #  0: save sequence to tmp_consensus -> 5

    state = 1
    #identifier = ""  # Currently not implemented.
    preMSA = []
    consensus = []
    for i, line in enumerate(infile):
        LOG.debug("Line %d: %s", i, line[0:-1])

        # CURRENTLY NOT IMPLEMENTED
        # if state == 0: #searching for sequenceMSA identifier
        #  tmp = pat_identifier.search(line)
        #  if tmp:
        #    state = 1
        #    identifier = tmp.group()
        #  continue
        # elif state == 1: #searching for TR boundaries (indices)

        if 1 == state:  # searching for repeat region coordinates
            search = pat_coordinates.search(line)
            if search:
                LOG.debug(" * (1->2) Found coordinates")
                state = 2
                region = RepeatRegion()
                region.begin = int(search.group(1))
                #region_end = search.group(2)
                short = False

        # searching for beginning of MSA & save sequence to tmpMSA-> 4
        elif state == 2:
            match = pattern_seq.match(line)
            if match and match.group(1) == str(region.begin):
                LOG.debug(" *(2->4) Found first row of first MSA repeat unit")
                state = 4
                if len(match.group(2).strip().split(" ")) > 1:
                    short = True
                    preMSA = match.group(2).strip().split(" ")
                    LOG.debug("Repeat unit is short; preMSA: %s", str(preMSA))
                else:
                    tmpMSA = [match.group(2).strip().split(" ")[0]]
                    LOG.debug(" tmpMSA: %s", str(tmpMSA))

        elif state == 3:  # new repeat unit: save sequence to tmpMSA -> 4
            match = pattern_seq.match(line)
            if match:
                LOG.debug(" *(3->5) Found first row of new repeat unit")
                state = 6
                if short:
                    preMSA += match.group(2).strip().split(" ")
                    LOG.debug("Repeat unit is short; preMSA: %s", str(preMSA))
                else:
                    tmpMSA.append(match.group(2).strip().split(" ")[0])
                    LOG.debug(" tmpMSA: %s", str(tmpMSA))
            # if end: save tmpMSA to preMSA; save tmp_consensus to consensus;
            # Yield repeat region -> 1
            if pat_statistics.search(line):
                LOG.debug(
                    " *(5->1) Encountered 'Statistics': No more repeats, yielding.")
                state = 1
                if not short:
                    preMSA.append("".join(tmpMSA))
                    consensus.append("".join(tmp_consensus))
                region.msa = getMSA(preMSA, consensus)
                if len(region.msa) >= 2:
                    yield region
                preMSA = []
                consensus = []

        elif state == 4:  # new repeat unit: save sequence to tmp_consensus -> 5
            match = pattern_seq.match(line)
            if match:
                LOG.debug(
                    " *(4->5) Found first consensus row of the repeat unit")
                state = 5
                if short:
                    consensus = match.group(2).strip().split(" ")
                    LOG.debug(
                        "Repeat unit is short;  consensus: %s",
                        str(consensus))
                else:
                    tmp_consensus = [match.group(2).strip().split(" ")[0]]
                    LOG.debug(" tmp_consensus: %s", str(tmp_consensus))

        elif state == 5:  # SEARCHING FOR MSA ROW
            # if sequence: save sequence to tmpMSA -> 6
            match = pattern_seq.match(line)
            if match:
                LOG.debug(" *(5->6) Found a MSA row")
                state = 6
                if short:
                    preMSA += match.group(2).strip().split(" ")
                    LOG.debug("Repeat unit is short; preMSA: %s", str(preMSA))
                else:
                    tmpMSA.append(match.group(2).strip().split(" ")[0])
                    LOG.debug(" tmpMSA: %s", str(tmpMSA))

            # if end: save tmpMSA to preMSA; save tmp_consensus to consensus;
            # Yield repeat region -> 1
            if pat_statistics.search(line):
                LOG.debug(
                    " *(5->1) Encountered 'Statistics': No more repeats, yielding.")
                state = 1
                if not short:
                    preMSA.append("".join(tmpMSA))
                    consensus.append("".join(tmp_consensus))
                region.msa = getMSA(preMSA, consensus)
                if len(region.msa) >= 2:
                    yield region
                preMSA = []
                consensus = []

        elif state == 6:  # new repeat unit? ## SEARCHING FOR CONSENSUS ROW
            match = pattern_seq.match(line)
            # 1: save tmp_consensus -> 3
            if match and match.group(1) == '1':
                LOG.debug(
                    " *(6->3) Found a consensus row of a new repeat unit")
                state = 3
                if short:  # NEEDS TO BE CODED
                    consensus += match.group(2).strip().split(" ")
                    LOG.debug(
                        "Repeat unit is short; consensus: %s",
                        str(consensus))
                else:
                    newMSA = tmpMSA.pop()
                    preMSA.append("".join(tmpMSA))
                    consensus.append("".join(tmp_consensus))
                    tmpMSA = [newMSA]
                    tmp_consensus = [match.group(2).strip().split(" ")[0]]

            # 0: save sequence to tmp_consensus -> 5
            elif match:
                LOG.debug(" *(6->5) Found a consensus row")
                state = 5
                tmp_consensus.append(match.group(2).strip().split(" ")[0])

            # YIELD
            # aha! there should have been a consensus sequence, but there is
            # not. Hence we are finished with this repeat!
            else:
                LOG.debug(' *(6->1) No consensus row: repeat finished')
                state = 1
                if short:
                    preMSA = preMSA[:-1]
                else:
                    preMSA.append(
                        "".join(
                            tmpMSA[
                                0:-
                                1]))  # The last tmpMSA entry was not a repeat unit
                    consensus.append("".join(tmp_consensus))
                LOG.debug(" preMSA: %s", str(preMSA))
                LOG.debug(" consensus: %s", str(consensus))
                region.msa = getMSA(preMSA, consensus)
                if len(region.msa) >= 2:
                    yield region
                preMSA = []
                consensus = []


def getMSA(sequenceMSA, consensusMSA):
    """ Derive the MSA from a strange combination of consensusMSA and sequenceMSA in TRF
    (Benson) txt.html output files

    Args:
        sequenceMSA (?):
        consensusMSA (?):

    Returns:
         msa (list of str): The multiple sequence alignment predicted by TRF.
    """

    msa = [""] * len(sequenceMSA)

    while consensusMSA:
        # CHECK for insertions
        insertion = 1
        while insertion and consensusMSA:
            insertion = 0
            for i_con in consensusMSA:
                if i_con and i_con[0] == "-":
                    insertion = 1
                    break
            # INCLUDE insertions into the msa
            if insertion:
                for i in range(len(consensusMSA)):
                    if consensusMSA[i] and consensusMSA[i][0] == "-":
                        msa[i] += sequenceMSA[i][0]
                        sequenceMSA[i] = sequenceMSA[i][1:]
                        consensusMSA[i] = consensusMSA[i][1:]
                    else:
                        msa[i] += "-"

        # CHECK for deletions and normal sequence
        if not consensusMSA[0]:
            break

        for i in range(len(consensusMSA)):
            # The last repeat unit can be shorter than the ones before
            if not sequenceMSA[i]:
                break

            msa[i] += sequenceMSA[i][0]
            sequenceMSA[i] = sequenceMSA[i][1:]
            consensusMSA[i] = consensusMSA[i][1:]

    return msa

# ############## HHrepID - Soeding ###########################################


def hhpredid_get_repeats(infile):
    """ Read repeats from a HHREPID standard output (stdout) file stream successively.

    Read repeats from a HHREPID standard output (stdout) file stream successively.
    Postcondition: infile points to EOF.

    Layout of HHREPID standard output::

        protein ::=
             begin"-"\d    "+"\d repeatUnit
           ( \d"-"\d    "+"\d repeatUnit )+

    Args:
        infile (file stream): File stream from HHREPID standard output.
        [Generated by e.g.: ./hhrepid_32 -i FASTAFILE -v 0 -d cal.hhm -o INFILE]

    Returns:
         (Repeat): A generator function is returned that, when called in a loop,
         yields one repeat per iteration.

    .. todo:: Layout HHREPID output syntax.
    """

    # find a part of a repeat unit and its first coordinate
    # minus or \minus?

    pattern_repeat_unit_count = re.compile("Repeats\s+(\d+)")
    pattern_seq = re.compile("[A-Z]+(\d+).*(\d+)\-.*\+[\d]+ ([\-a-zA-Z.]+)")

    # Our possible parser states:

    # state1: Find number of repeat units n
    # state2: Find first (partial) row of the MSA
    # state3: Find all other (partial) rows of the MSA

    region = None
    state = 1
    for i, line in enumerate(infile):
        LOG.debug("Line %d: %s", i, line[0:-1])

        if 1 == state:  # Find 'Repeats' marker of new repeat
            search = pattern_repeat_unit_count.search(line)
            if search:
                LOG.debug(" *(1->2) Found repeat")
                state = 2
                n = int(search.group(1))

        elif 2 == state:  # Find first (partial) row of the MSA
            search = pattern_seq.search(line)
            if search:
                LOG.debug(" *(2->3) Found first repeat unit (part)")
                state = 3
                region = RepeatRegion()
                region.begin = int(search.group(2))
                region.msa = [""] * n
                region.msa[int(search.group(1)) -
                           1] = search.group(3).replace('.', '-').upper()

        elif 3 == state:  # Find all other (partial) rows of the MSA
            search = pattern_seq.search(line)
            if search:
                LOG.debug(" *(3->3) Found other repeat unit (part)")
                region.msa[int(search.group(1)) -
                           1] += search.group(3).replace('.', '-').upper()
            else:
                search = pattern_repeat_unit_count.search(line)
                if search:
                    LOG.debug(" *(3->2) Yield Repeat, begin next")
                    state = 2
                    n = int(search.group(1))
                    if len(region.msa) >= 2:
                        yield region
                        region = None
                    else:
                        log.warning(
                            "HHPREDID: Msa too short %s", str(
                                region.msa))

    # Yield final repeat region.
    if not region is None:
        if len(region.msa) >= 2:
            yield region
        else:
            log.warning("HHPREDID: Msa too short %s", str(region.msa))

####################################### Phobos TRF  ######################


def phobos_get_repeats(infile):
    """ Read repeats from a PHOBOS output file stream successively.

    Read repeats from a PHOBOS output file stream successively.
    Postcondition: infile points to EOF.

    Args:
        infile (file stream): File stream from PHOBOS output.

    Returns:
         (Repeat): A generator function is returned that, when called in a loop,
         yields one repeat per iteration.

    .. todo:: Show PHOBOS output syntax.
    """

    pattern_begin = re.compile("(\d+) :\s+\d")
    pattern_seq = re.compile("([\-ACGT]+)")

    # Our possible parser states:
    #
    # state 1: Find TR begin
    # state 2: Find first repeat unit
    # state 3: Find repeat units

    state = 1
    for i, line in enumerate(infile):
        LOG.debug("Line %d: %s", i, line[0:-1])
        if 1 == state:  # Find TR offset
            search = pattern_begin.search(line)
            if search and search.groups()[0] != None:
                LOG.debug(" *(1->2) Found tandem repeat begin")
                state = 2
                region = RepeatRegion()
                region.begin = int(search.groups()[0])
                region.msa = []

        elif 2 == state:  # Find all other repeat units
            match = pattern_seq.search(line)
            if match and match.groups()[0] != None:
                LOG.debug(" *(2->3) Found first repeat unit")
                region.msa.append(match.groups()[0])
                state = 3

        elif 3 == state:  # Find all other repeat units
            match = pattern_seq.search(line)
            if match and match.groups()[0] != None:
                LOG.debug(" *(3->3) Found a repeat unit")
                region.msa.append(match.groups()[0])
            else:
                LOG.debug(" *(3->1) repeat region finished, yielding.")
                state = 1
                if len(region.msa) >= 2:
                    yield region
                else:
                    log.warning("phobos: Msa too short %s", str(region.msa))



import pytest

from tral.repeat import repeat
from tral.repeat import repeat_io


TEST_MSA = ['HPFGFV-------------AVPTKNP-DGTMNLMNWECAIPGKKGTPWEGGLFKLRMLFKDDYPS---SPPKCKFEPPLFHPNV', 'YPSGTVCLsileedkdwrpAITIKQIlLGIQELLN-E---PNIQ-DPAQAEAYTIYCQNRVEYEKrvrAQAK-KFAP-------']
TEST_BEGIN = 0

notfixed = pytest.mark.notfixed

@notfixed
@pytest.mark.no_external_software_required
def test_repeat_score():

    ### MAKE THIS TEST A TEST!
    Q,eqFreq,alphabet = repeat_io.loadModel()
    indelRatePerSite = 0.001

    myTR = repeat.Repeat(msa = TEST_MSA, begin = TEST_BEGIN)
    myTR.deleteInsertionColumns()

    print(loglikelihood_gaps_starphylogeny_zipfian(t = 1, tandem_repeat = myTR))
    print(optimisation(function = loglikelihood_substitution, args = [Q,eqFreq,alphabet,myTR]))
    print(optimisation(function = loglikelihood_substitutions_gaps, args = [[Q,eqFreq,alphabet,myTR],[myTR,indelRatePerSite]]))

    assert 1 == 2



'''
Implementation of the workflow used in :
Schaper,E. et al. (2014) Deep conservation of human protein tandem repeats within the eukaryotes. Molecular Biology and Evolution. 31, 1132–1148 .
'''

import logging
import logging.config
import os

from tral.paths import config_file, PACKAGE_DIRECTORY

from tral.sequence import repeat_detection_run, sequence
from tral.hmm import hmm

logging.config.fileConfig(config_file("logging.ini"))
log = logging.getLogger('root')

TEST_FAA_FILE_MBE_2014 = os.path.join(
    PACKAGE_DIRECTORY,
    "test",
    "P51610.fasta")
TEST_HMM_FILES_MBE_2014 = [
    os.path.join(
        PACKAGE_DIRECTORY,
        "test",
        "Kelch_1.hmm"),
    os.path.join(
        PACKAGE_DIRECTORY,
        "test",
        "Kelch_2.hmm")]
TEST_SCORE_MBE_2014 = "phylo_gap01_ignore_trailing_gaps_and_coherent_deletions"


def path():
    """Return the path to the test data files.
    """
    return os.path.join(os.path.abspath('.'), 'tandemrepeats', 'test')


def sample_MBE_2014_pipeline():
    # The Schaper et al. (MBE, 2014) pipeline is tested on a single sequence.

    test_lSeq = sequence.Sequence.create(
        os.path.join(
            path(),
            TEST_FAA_FILE_MBE_2014),
        input_format="fasta")
    test_seq = test_lSeq[0]

    # Information on sequence domains (here: Pfam) in this sequence are added.
    test_pfam_hmm = [
        hmm.HMM.create(
            input_format="hmmer",
            file=os.path.join(
                path(),
                i)) for i in TEST_HMM_FILES_MBE_2014]

    # The sequence is searched for tandem repetitions of the Pfam domain in
    # the sequence
    test_pfam_list = test_seq.detect(lHMM=test_pfam_hmm)
    assert len(test_pfam_list.repeats) == 2

    # Pfam TRs with n_effective < 3.5 are discarded.
    test_pfam_list = test_pfam_list.filter("attribute", "n_effective", "min", 3.5)
    assert len(test_pfam_list.repeats) == 2

    # de novo detection methods (Trust, T-reks, Xstream, HHrepID) are used to search the
    # INSERT OWN PARAMTERS USING: test_denovo_list = test_seq.detect(denovo =
    # True, **TEST_DENOVO_PARAMETERS)
    test_denovo_list = test_seq.detect(denovo=True)
    # When Trust is part of the detectors, the number of found repeats may
    # differ between runs...
    assert len(test_denovo_list.repeats) == 10

    # De novo TRs with dTR_units (divergence) > 0.8; n_effective < 2.5; l < 10 or
    # pvalue "phylo_gap01_ignore_trailing_gaps_and_coherent_deletions" > 0.01
    # are discarded.
    test_denovo_list = test_denovo_list.filter(
        "pvalue",
        TEST_SCORE_MBE_2014,
        0.01)
    assert len(test_denovo_list.repeats) == 10
    test_denovo_list = test_denovo_list.filter(
        "divergence",
        TEST_SCORE_MBE_2014,
        0.8)
    assert len(test_denovo_list.repeats) == 10
    test_denovo_list = test_denovo_list.filter("attribute", "n_effective", "min", 2.5)
    assert len(test_denovo_list.repeats) == 5
    test_denovo_list = test_denovo_list.filter("attribute", "l", "min", 10)
    assert len(test_denovo_list.repeats) == 2

    # De novo TRs were remastered with HMM
    test_denovo_hmm = [
        hmm.HMM.create(
            repeat=iTR) for iTR in test_denovo_list.repeats]
    test_denovo_list_remastered = test_seq.detect(lHMM=test_denovo_hmm)
    assert len(test_denovo_list_remastered.repeats) == 2

    # pvalue "phylo_gap01_ignore_trailing_gaps_and_coherent_deletions" > 0.1
    # are discarded.
    test_denovo_list_remastered = test_denovo_list_remastered.filter(
        "pvalue",
        TEST_SCORE_MBE_2014,
        0.1)

    # De novo TRs were filtered (n_effective < 3.5 are discarded.)
    test_denovo_list_remastered = test_denovo_list_remastered.filter(
        "attribute",
        "n_effective",
        "min",
        3.5)
    assert len(test_denovo_list_remastered.repeats) == 2

    # De novo TRs overlapping with a Pfam TR were filtered
    test_denovo_list_remastered = test_denovo_list_remastered.filter(
        "none_overlapping_fixed_repeats",
        test_pfam_list,
        "shared_char")
    assert len(test_denovo_list_remastered.repeats) == 2

    # Remaining De novo TRs were clustered for overlap (common ancestry). Only best =
    # lowest p-Value and lowest divergence were retained.
    test_denovo_list_remastered = test_denovo_list_remastered.filter(
        "none_overlapping", ["common_ancestry"], {
            "pvalue": TEST_SCORE_MBE_2014, "divergence": TEST_SCORE_MBE_2014})
    assert len(test_denovo_list_remastered.repeats) == 1

    # Merge remaining set of de novo and Pfam TRs.
    test_entire_set = test_pfam_list + test_denovo_list_remastered
    assert len(test_entire_set.repeats) == 3

    # Write result set of Pfam TRs
    #test_entire_set.write(format = "tsv,...")


if __name__ == "__main__":

    sample_MBE_2014_pipeline()



# (C) 2012-2015 Elke Schaper
# (C) 2014 Julia Pecerska

"""
    :synopsis: A cyclic hidden Markov model that describes sequence tandem repeats.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb.sib.ch>
"""

import datetime
import logging
import numpy as np
import os
import pickle
import shutil
import subprocess
import tempfile

from tral.hmm import hmm_io, hmm_viterbi
from tral.repeat.repeat import Repeat
from tral import configuration

LOG = logging.getLogger(__name__)

CONFIG_GENERAL = configuration.Configuration.instance().config
CONFIG = CONFIG_GENERAL["hmm"]

################################### HMM class ############################


class HMM:

    """ Sequence profile hidden Markov models (HMMs) are used to model genomic
    sequence containing a tandem repeat.

    Note:
        The HMM implemented here is described in detail in

        Schaper, E., Gascuel, O. & Anisimova, M. Deep conservation of human
        protein tandem repeats within the eukaryotes.
        Molecular Biology and Evolution (2014).


        All notations used here are based on the notations introduced in

        Eddy, S. R. Profile hidden Markov models.
        Bioinformatics 14, 755–763 (1998).

    The HMM contains flanking states ("N", "C"), match states ("M1", "M2", ...)
    and insertion states ("I1", "I2", ..). As deletion states are non-emitting,
    they are merged with the other states.

    The HMM state "N" emits the sequence before the tandem repeat.
    The HMM state "C" emits the residual sequence after the tandem repeat.
    The HMM states "M1", "M2", ... can emit sequence within the tandem repeat.
    "M1" describes the first amino acid/nucleotide in the repeat unit,
    "M2" the second, and so on]. The HMM states "I1", "I2", can emit sequence
    representing insertions within the tandem repeat.

    The transition probabilities between states are stored in ``p_t``.
    The emission probabilities of every state are stored in ``p_e``.
    The probabilities for any state to be the null state are stored in ``p_0``.

    The structure and probabilities defining standard sequence profile HMMs can
    be

    * taken from HMM databases such as PFAM,
    * calculated from sequence alignments (here: an alignment of TR units) using for example the HMMER suite,
    * defined in house.

    All three ways are implemented here.

    The HMM is used to detect the maximum likelihood occurrence of a TR modeled
    by the HMM on a sequence. The Viterbi algorithm is implemented to do this
    detection.

    Attributes:
        id (str): The id describes the origin of the HMM.
                    E.g., PFAMID "PF00069".
        states (list of str): The names of all states that the HMM contains.
                              E.g. ["N", "M1", "I1", "M2", "I2", "C"]
        insertion_states (list of str): The names of all insertion states that
                                        the HMM contains. E.g. ["I1", "I2"]
        match_states (list of str): The names of all match states that the HMM
                                    contains. E.g. ["M1", "M2"]
        l_effective (int): The number of states in the HMM.
                           E.g. l_effective = 4
        alphabet (list of str): The letters that the HMM states can emit.
                                E.g. all amino acids in the LG matrix:
                                ["A", "C", ...]
        p_t (dict of dict of float): A dictionary
            {state0: {state1: transition probability, ...}, ...}
            between all `states` as log10.
        p_0 (dict of dict of float): A dictionary of the null probabilities to
                                    be in any of the `states`:
                                    {`states[0]`: 0.1, `states[1]`: 0.2, ...}
        p_e (dict of dict of float): A dictionary of emission probabilities for
                every state and every letter in `alphabet`:
                {`states[0]`: {`alphabet[0]`: emission probability, ...}, ...}
        hmmer (dict of dict of float): A dictionary of all information
                                       extracted from a Hmmer model.
    """

    def __str__(self):
        """
            Create string for HMM instance.
        """
        try:
            tmp = [self.p_e[i] for i in self.match_states]
            tmp = "".join([max(i.keys(), key=(lambda key: i[key]))
                           for i in tmp])
            hmm = "cpHMM ID: {}\ncpHMM length: {}\n".format(
                self.id,
                self.l_effective)
            hmm += "Most likely motif: {}".format(tmp)
        except:
            hmm = "<HMM instance>"
            LOG.warning("Could not create string of HMM instance.")

        return hmm

    # Use functions defined in other methods as class functions.
    # Static function (declare as @static ?):
    read = hmm_io.read

    # Function that take the HMM instance as argument:
    def viterbi(self, *args):
        return hmm_viterbi.viterbi(self, *args)

    def __init__(self, hmmer_probabilities):
        """ HMM class __init_ module.

        __init__ takes HMM parameters (including the alphabet, emission
        probabilities and transition probabilities) as input, and assigns
        them to class attributes.

        Args:
            hmmer_probabilities (dict): A dictionary with HMM parameters.
        """
        self.hmmer = hmmer_probabilities
        self.id = hmmer_probabilities['id']
        self.alphabet = self.hmmer['letters']

        self.l_effective = max([int(key) for key \
                               in hmmer_probabilities.keys() if key.isdigit()])

        # Initialise all HMM states to default value (e.g. transition to or
        # from terminal state all have the same cost 0).
        self.initialise_HMM_structure(self.l_effective)

        d_translate_states = {
            i: i[
                1:] for i in self.match_states +
            self.insertion_states}
        for i in self.terminal_states:
            d_translate_states[i] = "COMPO"

        # Store null model emission probabilities for ["N","C"]
        # from hmmer_probabilities["COMPO"]
        # Store match_state and insertion_state emission probabilities for
        # ["M0",...] from hmmer_probabilities["0"],...
        for i_state in self.match_states:
            i_emission_probabilities = self.hmmer[
                d_translate_states[i_state]]['emissions'][:len(self.alphabet)]
            self.set_emission_probability_hmmer3(
                i_state,
                i_emission_probabilities)
        for i_state in (self.insertion_states + self.terminal_states):
            i_emission_probabilities = \
                self.hmmer[d_translate_states[i_state]] \
                ['insertion_emissions'][:len(self.alphabet)]
            self.set_emission_probability_hmmer3(
                i_state,
                i_emission_probabilities)

        # Store transition probabilities
        # (m->m    m->i    m->d    i->m    i->i    d->m    d->d):
        # First: All state defined in HMMer model
        for i, i_state in enumerate(self.match_states):
            i_transition_probabilities = self.hmmer[
                d_translate_states[i_state]]['transition']
            # Completing the circle of the HMM.
            if i == self.l_effective - 1:
                i_transition_probabilities = self.set_circle_transition_probability_hmmer3(
                    i_transition_probabilities,
                    d_translate_states[i_state])
            LOG.debug("i_state: %s", i_state)
            LOG.debug(
                "set_circle_transition_probability_hmmer3: %s",
                i_transition_probabilities)
            self.set_transition_probability_hmmer3(
                i,
                i_transition_probabilities)

        LOG.debug("p_t before deletion: %s", self.p_t)

        # Translate deletion states into direct transitions from match to match
        # states.
        if self.l_effective > 1:
            p_t_d = {}
            for i, i_match_state in enumerate(self.match_states):
                p_t_d[i_match_state] = self.get_direct_transition_probabilities_for_deletions(
                    i,
                    i_match_state)
            for i_match_state in self.match_states:
                for i_goal_state, iP in p_t_d[i_match_state].items():
                    self.p_t[i_match_state][i_goal_state] = iP

        LOG.debug("p_t after deletion: %s", self.p_t)

        # As it is never likely to got from a terminal state to a deletion
        #  tate or vice versa, this transition probabilities can be ignored.
        # Thus, you may advance and:
        # Delete everything you ever knew about deletion states.

        self.states = [self.terminal_states[
            0]] + self.match_states + self.insertion_states + [self.terminal_states[1]]
        self.p_t = {i_state: {i: j for i, j in self.p_t[
            i_state].items() if not i in self.deletion_states} for i_state in self.states}
        self.p_0 = {
            i: j for i,
            j in self.p_0.items() if i not in self.deletion_states}

        del self.deletion_states

    def create(input_format, file=None, repeat=None):
        """ Creates a HMM instance from 2 possible input formats.

        A `HMM` instance is created from one of the two possible inputs:

        * HMMER3 model
        * ``Repeat`` instance

        Args:
            input_format (str): The file format of the input file
            file (str): Path to the file containing the HMM parameters
                encoded in the HMMER3 format.
            repeat (Repeat): A Repeat object with an MSA that can be
                transformed into an HMM.

        Returns:
            `HMM`: An initialized instance of the `HMM` class.

        Raises:
            Exception: if the HMMER3 file provided does not exist.
            Exception: if the repeat value provided is not an instance of the
                Repeat class.
            Exception: if no parameters are provided to create the HMM from.

        .. warning:: Will initialize all HMM's in a file but will only return
            the first one.
        .. todo:: Fix the previous warning (agree on the way how it should be
            done first)

        """

        if input_format == 'hmmer':
            # Read only first element from HMMER3 file.
            if not os.path.exists(file):
                raise Exception('HMMER3 file does not exist.')
            hmmer_probabilities = next(HMM.read(file))
        elif input_format == 'pickle':
            with open(file, 'rb') as fh:
                return pickle.load(fh)
        elif input_format == 'repeat':
            if not isinstance(repeat, Repeat):
                raise Exception('The repeat value is not a valid instance of '
                                'the Repeat class.')
            hmmer_probabilities = HMM.create_from_repeat(repeat)
        else:
            raise Exception("Unknown input format: {}.".format(input_format))

        LOG.debug(hmmer_probabilities)
        return HMM(hmmer_probabilities)

    def create_from_repeat(tandem_repeat, hmm_copy_path=None,
                           hmm_copy_id=None):
        """ Get HMM parameters (including the alphabet, emission probabilities
        and transition probabilities) from a tandem repeat.

        An HMM is created from ``tandem_repeat`` using :command:`hmmbuild`
        and is saven in the HMMER3 file format.
        Next, HMM parameters are retrieved from the HMMER3 file and returned.

        Args:
            tandem_repeat (TR):  A Repeat class instance of the TR to be
                transformed into an HMM.
            hmm_copy_path (str): Path to where a copy of the created HMM
                will be stored.
                If None, no copies are saved.
            hmm_copy_id (str): HMM id which will serve as the filename in case
                a copy should be saved.
                If None and ``hmm_copy_path`` is specified the name is
                generated from current date and time combination.

        Returns:
            HMM parameters (dict): For a sample structure please refer to
            :py:hmm_io.read:.

        .. todo:: Make sure the id is correctly read in.

        """
        timestamp = datetime.datetime.now()
        tmp_id = timestamp.strftime('%Y:%m:%d:%H:%M:%S:%f')

        # Create a temporary directory
        tmp_dir = tempfile.mkdtemp()

        # Save TR as Stockholm file
        stockholm_file = os.path.join(tmp_dir, tmp_id + ".sto")

        # O repeat_io.save_repeat_stockholm(tandem_repeat.msaD, stockholm_file)
        tandem_repeat.write(file=stockholm_file, file_format="stockholm")

        # Run HMMbuild to build a HMM model, and read model
        p = subprocess.Popen([CONFIG["hmmbuild"], "--amino", tmp_id + ".hmm",
                              tmp_id + ".sto"],
                             stdout=subprocess.PIPE, stderr=None, cwd=tmp_dir)
        p.wait()

        hmm_file = os.path.join(tmp_dir, tmp_id + ".hmm")
        if hmm_copy_path:
            if not os.path.exists(hmm_copy_path):
                LOG.critical('Specified path for the file copy does not '
                             'exist, not saving copy: %s.', hmm_copy_path)
            else:
                if hmm_copy_id:
                    shutil.copy(hmm_file, os.path.join(hmm_copy_path,
                                                       hmm_copy_id + ".hmm"))
                else:
                    shutil.copy(hmm_file, hmm_copy_path)

        hmmer_probabilities = next(
            HMM.read(
                hmm_filename=hmm_file,
                id=hmm_copy_id))

        shutil.rmtree(tmp_dir)

        return hmmer_probabilities

    def write(self, file, output_format, *args):
        """ Write ``HMM`` to file.

        Write ``HMM`` to file. Currently, only pickle is implemented.

        Args:
            file (str): Path to input file
            output_format (str):  Either "fasta", "pickle" or "stockholm"

        .. todo:: Write checks for ``format`` and ``file``.

        """

        if output_format == 'pickle':
            with open(file, 'wb') as fh:
                pickle.dump(self, fh)
        else:
            raise Exception('output_format is unknown: {}'.format(output_format))

    def hmm_example(self):
        """ .. todo:: Is this method needed?
        """
        #states = ["N", "B", "M1", "M2", "M3", "E", "C"]
        self.states = ["N", "M1", "M2", "C"]

        # Initialisation
        self.p_t = {iS: {iS2: 0 for iS2 in self.states}
                    for iS in self.states[:-1]}
        self.p_0 = {iS: 1 / 3 for iS in self.states}

        # Transition probabilities
        # Feed Values to p_t
        self.p_t["N"] = {"N": 0.5, "M1": 0.5}
        self.p_t["M1"] = {"M2": 0.5, "C": 0.5}
        self.p_t["M2"] = {"M1": 0.5, "C": 0.5}
        self.p_t["C"] = {"C": 1}

        # emissions
        self.emissions = ["A", "C", "G", "T"]

        # emission probabilities
        self.p_e = {iS: {iE: 0.25 for iE in self.emissions}
                    for iS in self.states}
        self.p_e['M1'] = {"A": 0.9, "C": 0.025, "G": 0.025, "T": 0.025}
        self.p_e['M2'] = {"A": 0.025, "C": 0.9, "G": 0.025, "T": 0.025}

    def initialise_HMM_structure(self, l_effective):
        """ Initialise the HMM with all states and standard values for
            transition and emission probabilities.

        Initialise all states
        Set transition probabilities to None/0
        Set initial probability for all states equal.

        Args:
          l_effective (int): The number of match states in the HMM.

        """

        # Build a HMM from these likelihoods.
        self.match_states = ["M{0}".format(str(i + 1)) for i in range(l_effective)]
        self.insertion_states = ["I{0}".format(str(i + 1)) for i in range(l_effective)]
        self.deletion_states = ["D{0}".format(str(i + 1)) for i in range(l_effective)]
        self.terminal_states = ["N", "C"]
        self.states = [self.terminal_states[0]] + self.match_states + \
            self.insertion_states + self.deletion_states + [self.terminal_states[1]]
        LOG.debug("HMM states: %s", self.states)

        # Initialisation
        # The transition probability is initially set to None for all states.
        self.p_t = {iS: {iS2: None for iS2 in self.states}
                    for iS in self.states}
        # The initial probability is equal in all states. (np.log10(1)=0)
        self.p_0 = {iS: 0 for iS in self.states}

        # Transition probabilities

        # First, mark which transitions are not penalised,
        # that is bear a 'probability of np.log10(1)=0':
        # - Transitions from "N" to any "M_k"
        # - Transitions from any "M_k" to "C"
        # - Transitions from "M_k" to "M_k+1" (for k`n) and "M_k" to "M_0" (for k==n)
        # - Transitions from "N" to "N"
        # - Transitions from "C" to "C"
        for i in range(l_effective):
            self.p_t["N"]["M{0}".format(i % l_effective + 1)] = 0
            self.p_t["M{0}".format(i % l_effective + 1)]["C"] = 0
            self.p_t["M{0}".format(i % l_effective + 1)] \
                    ["M{0}".format((i + 1) % l_effective + 1)] = 0
        self.p_t["N"]["N"] = 0
        self.p_t["C"]["C"] = 0

        self.p_e = {}

    def set_emission_probability_hmmer3(self, state, l_emission_probabilities):
        """ Convert and set emission probabilities from HMMER3 file to class
            attribute.

        Set p_e of states to `l_emission_probabilities` for `state` given
        `self.alphabet`.
        In HMMER3 data, emission probabilities are -ln(p).

        It is e.g.::

            self.alphabet =
            ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q',
             'R', 'S', 'T', 'V', 'W', 'Y']

        Return log10(p), that is convert between the two. Conversion:
        p_Local = - p_HMM * log10(e)

        Args:
            state (str): The state of the HMM. E.g. the fourth match state 'M4'.
            l_emission_probabilities (list of float): E.g.
                ['3.27687', '2.31397', '3.32252', '3.12746', '2.89175',
                '3.34719', '2.28730', '3.54139', '2.53154', '2.64774',
                '3.75733', '3.23860', '3.57894', '3.26290', '2.66343',
                '2.61544', '2.91770', '3.26739', '5.09378', '3.16816']

        """

        l_emission_probabilities = \
            [-(i) * np.log10(np.exp(1)) for i in l_emission_probabilities]
        self.p_e[state] = \
            {iL: iEP for iL, iEP in \
            zip(self.alphabet, l_emission_probabilities)}

    def set_circle_transition_probability_hmmer3(
            self,
            l_transition_probabilities,
            final_state):
        """ Calculate transition probabilites between states that exist due to
            the circularity of the tandem repeat HMM.

        In HMMER3 models, no transition in circle from the last match state to
        the first match state are included.

        Values need to be found for the following transitions:
        m->m    m->i    m->d    i->m    i->i    d->m    d->d

        Compare ftp://selab.janelia.org/pub/software/hmmer3/3.0/Userguide.pdf
        pp.84-85.

        Input: -ln(p) Output: Also -ln(p')

        When the HMM only has a single match state, i.e. l_effective == 1, the
        final match state equals the only match state, and no transition
        probabilities can be calculated via averaging over all other states.
        Therefore, the its transition probabilities remain unchanged.

        Args:
          l_transition_probabilities (list of float): The first parameter.
          final_state (str): The second parameter.

        Returns:
          list of float: transition probabilities

        """

        if self.l_effective == 1:
            return l_transition_probabilities

        # In self.hmmer, besides the probabilities for states 1, 2, 3, ... data
        # for "COMPO", 'letters', and 'id' are saved and should be ignored in
        # the following.
        # Also, we ignore the transition probabilities for final_state, as our
        # goal here is to recalculate these based on the averages of the
        # transition probabilities from all other states.
        skip_states = ['COMPO', 'letters', final_state, 'id']

        # M->M, M->I, M->D: Use an average from all other transitions from
        # match states.
        LOG.debug("self.hmmer.keys(): %s", self.hmmer.items())
        LOG.debug("self.hmmer.items(): %s", self.hmmer.items())
        mean_M_M = np.mean([np.exp(-iP["transition"][0])
                            for i_state, iP in self.hmmer.items() if i_state not in skip_states])
        mean_M_I = np.mean([np.exp(-iP["transition"][1])
                            for i_state, iP in self.hmmer.items() if i_state not in skip_states])
        mean_M_D = np.mean([np.exp(-iP["transition"][2])
                            for i_state, iP in self.hmmer.items() if i_state not in skip_states])

        sum_p = np.sum([mean_M_M, mean_M_I, mean_M_D])
        l_transition_probabilities[
            :3] = [-np.log(i / sum_p) for i in [mean_M_M, mean_M_I, mean_M_D]]

        # I->M and I->I: Use an average from all other transitions from
        # insertion states.
        mean_I_M = np.mean([np.exp(-iP["transition"][3])
                            for i_state, iP in self.hmmer.items() if i_state not in skip_states])
        mean_I_I = np.mean([np.exp(-iP["transition"][4])
                            for i_state, iP in self.hmmer.items() if i_state not in skip_states])
        sum_p = np.sum([mean_I_M, mean_I_I])
        l_transition_probabilities[
            3:5] = [-np.log(i / sum_p) for i in [mean_I_M, mean_I_I]]

        # D->M and D->D: Use an average from all other transitions from
        # deletion states.
        mean_D_M = np.mean([np.exp(-iP["transition"][5])
                            for i_state, iP in self.hmmer.items() if i_state not in skip_states])
        mean_D_D = np.mean([np.exp(-iP["transition"][6])
                            for i_state, iP in self.hmmer.items() if i_state not in skip_states])
        sum_p = np.sum([mean_D_M, mean_D_D])
        l_transition_probabilities[
            5:] = [-np.log(i / sum_p) for i in [mean_D_M, mean_D_D]]

        return l_transition_probabilities

    def set_transition_probability_hmmer3(
            self,
            state_index,
            l_transition_probabilities):
        """ Convert and assign transition probabilities from a HMMER3 file to
            HMM attribute `p_t`.


        Convert and assign transition probabilities from a HMMER3 file to
        HMM attribute `p_t`.
        Set ``p_t`` of states to ``l_transition_probabilities`` for the
        ``state_index`` th state.

        In HMMER3 data, transition probabilities are -ln(p).
        Return log10(p), i.d. convert between the two.
        Conversion: p_Local = - p_HMM * log10(e)

        Args:
            state_index (int): E.g. 4
            l_transition_probabilities (list of float): A list of transition
                probabilities as found in a single row for a certain state in
                HMMER3 files. E.g. ['0.00021', '8.85487', '9.57722', '0.61958',
                '0.77255', '0.48576', '0.95510']
        """

        l_transition_probabilities = [-(i) * np.log10(np.exp(1)) for i \
                                      in l_transition_probabilities]

        # Match state state_index -> Match state state_index + 1
        iM = self.match_states[state_index]
        iM_1 = self.match_states[(state_index + 1) % self.l_effective]
        self.p_t[iM][iM_1] = l_transition_probabilities[0]

        # Match state state_index -> Insertion state state_index
        self.p_t[iM][
            self.insertion_states[state_index]] = l_transition_probabilities[1]

        # Match state state_index -> Deletion state state_index + 1
        iD_1 = self.deletion_states[(state_index + 1) % self.l_effective]
        self.p_t[iM][iD_1] = l_transition_probabilities[2]

        # Insertion state state_index ->  Match state state_index + 1
        iI = self.insertion_states[state_index]
        self.p_t[iI][iM_1] = l_transition_probabilities[3]

        # Insertion state state_index ->  Insertion state state_index
        iI = self.insertion_states[state_index]
        self.p_t[iI][iI] = l_transition_probabilities[4]

        # Deletion state state_index -> Match state state_index + 1
        iD = self.deletion_states[state_index]
        self.p_t[iD][iM_1] = l_transition_probabilities[5]

        # Deletion state state_index -> Deletion state state_index + 1
        self.p_t[iD][iD_1] = l_transition_probabilities[6]

    def get_direct_transition_probabilities_for_deletions(
            self,
            state_index,
            state):
        """ Calculate updated transition probabilities for HMMs without
            deletion states.

        Calculate updated transition probabilities for HMMs without deletion
        states. Deletion states are non-emitting states. Therefore, they can be
        merged into the remaining emitting states. E.g. M -> D -> I would be
        equivalent to M -> I. The total transition probability from M to I
        would then be the product of the transition probabilities from M to D
        and from D to I (equivalent to the sum of the log10 probabilities.).

        This method translates all deletions following `state` with
        `state_index` into direct match state to match state transitions, and
        returns the new transition probabilities.

        Args:
            state_index (int): The first parameter.
            state (str): The second parameter.

        Returns:
            {state1: float}: The transition probability from `state` with
            `state_index` to `state1` is calculated for all (?) possible
            `state1`.
        """

        transition = {}
        p_deletion_state = self.p_t[state][
            self.deletion_states[
                (state_index + 1) %
                self.l_effective]]
        for i_state_index in range(1, self.l_effective):
            i_deleted_state_index = (state_index + i_state_index) % self.l_effective
            i_deleted_state = self.deletion_states[i_deleted_state_index]
            i_goal_state_index = (state_index + i_state_index + 1) % self.l_effective
            i_goal_state = self.match_states[i_goal_state_index]
            transition[i_goal_state] = p_deletion_state + \
                self.p_t[i_deleted_state][i_goal_state]
            p_deletion_state += self.p_t[i_deleted_state][
                self.deletion_states[i_goal_state_index]]

        return transition


############################### External parameters ######################

def hmmer3_emission_probabilities(hmmer_probabilities, letters, l_match):
    '''
    Get emission probabilities from hmmer3 hmm file.
    In hmm file, emission probabilities are -ln(p).
    Return log10(p), i.d. convert between the two.
    Conversion: p_Local = - p_HMM * log10(e)

    Parameters (e.g.)::

        letters = ['A', 'C', 'E', 'D', 'G', 'F', 'I', 'H', 'K', 'M', 'L', 'N',
                   'Q', 'P', 'S', 'R', 'T', 'W', 'V', 'Y']
        l_match = ['M'+str(i) for i in range(24)]

    Return format (pseudo code):
    [{iA: np.log10(p(iA,iM)) for iA in alphabet.keys()} for iM in l_match]

    '''

    # Test: Is the number of match states in both models equal?
    if not len(hmmer_probabilities.keys()) - 2 == len(l_match):
        print('Match states HMMER: {0} Match states local: {1}'.format(
            len(hmmer_probabilities.keys()) - 2, len(l_match)))
        raise ValueError('The number of match states in HMMer model and local'
                         'model does not match')

    # Test: Are all `letters` represented in the HMMER HMM?
    if any(iL not in hmmer_probabilities['letters'] for iL in letters):
        missing = [
            iL for iL in letters if iL not in hmmer_probabilities['letters']]
        print('Missing representation in Hmmer File: {0}'.format(missing))
        raise ValueError('Some letters in the local HMM are not represented',
                         'in the HMMER HMM.')

    return [{iL: -(iP) * np.log10(np.exp(1)) for iL,
             iP in zip(hmmer_probabilities['letters'],
                       data['emissions']) if iL in letters} for key,
            data in hmmer_probabilities.items() if key not in ['letters',
                                                               'COMPO']]



# (C) 2015 Elke Schaper

"""
    :synopsis: The Sequence Class.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import logging
import pickle
import re

from tral import configuration
from tral.repeat import repeat, repeat_align
from tral.repeat_list import repeat_list
from tral.hmm import hmm, hmm_viterbi
from tral.sequence import repeat_detection_run, sequence_io

CONFIG = configuration.Configuration.instance().config
LOG = logging.getLogger(__name__)


class Sequence:

    """ A ``Sequence`` describes either a protein or a DNA sequence.

    ``Sequence`` contains methods that act on single sequences, for example:

    *   tandem repeats

    Attributes:
        seq (str): The sequence.
        seq_standard_aa (str): The sequence with standard amino acids only
    """

    def __init__(self, seq, name=None):

        if not isinstance(seq, str):
            raise Exception('The seq value is not a String')
        self.seq = seq.upper()
        for i in self.seq:
            if i not in CONFIG['all_amino_acids']:
                raise Exception("{} is not in CONFIG['all_amino_acids']: {}"
                                .format(i, CONFIG['all_amino_acids']))

        self.seq_standard_aa = repeat.standardize(self.seq)

        if name:
            self.name = name

        self.d_annotations = {}
        self.d_repeatlist = {}

    def create(file, input_format):
        """ Create sequence(s) from file.

        Create sequence(s) from file.

        Args:
            file (str): Path to input file
            format (str):  Either "fasta" or "pickle"

        .. todo:: Write checks for ``format`` and ``file``.
        """

        if input_format == 'fasta':
            l_seq = sequence_io.read_fasta(file)
            return [Sequence(iSeq, iID) for iSeq, iID in l_seq]
        if input_format == 'pickle':
            with open(file, 'rb') as fh:
                return pickle.load(fh)
        else:
            raise Exception("Input format {} is not implemented for"
                            "sequence.create()".format(input_format))

    def write(self, file, file_format):
        """ Write sequence to file.

        Write sequence to file using one of two formats.

        Args:
            file (str): Path to output file
            format (str):  Either "fasta" or "pickle"

        .. todo:: Write checks for ``format`` and ``file``.

        """

        if file_format == 'fasta':
            sequence_io.write(self.seq, file)
        elif file_format == 'pickle':
            with open(file, 'wb') as fh:
                pickle.dump(self, fh)
        else:
            raise Exception("Output format {} is not implemented for",
                            "sequence.write()".format(file_format))

    def detect(self, lHMM=None, denovo=None, **kwargs):
        """ Detects tandem repeats on ``self.seq`` from 2 possible sources.

        A list of ``Repeat`` instances is created for tandem repeat detections
        on the sequence from two possible sources:

        * Sequence profile hidden Markov models ``HMM``
        * de novo detection algorithms.

        Args:
            hmm (HMM): A list of ``HMM`` instances.
            denovo (bool): boolean
            *kwargs: Parameters fed to denovo TR prediction and/or Repeat
                instantiation. E.g. ``repeat = {"calc_score": True}``

        Returns:
            A ``RepeatList`` instance
        """

        if lHMM:
            if not isinstance(lHMM, list):
                raise Exception('The lHMM value is not a list.')
            for iHMM in lHMM:
                if not isinstance(iHMM, hmm.HMM):
                    raise Exception('At least one list element in the lHMM'
                                    'value is not a valid instance of the HMM'
                                    'class.')

            repeats = []
            for iHMM in lHMM:
                # Detect TRs on self.seq with hmm using the Viterbi algorithm.
                most_likely_path = iHMM.viterbi(self.seq)
                LOG.debug("most_likely_path: {}".format(most_likely_path))
                if not most_likely_path:
                    continue
                unaligned_msa = hmm_viterbi.hmm_path_to_non_aligned_tandem_repeat_units(
                    self.seq,
                    most_likely_path,
                    iHMM.l_effective)
                if len(unaligned_msa) > 1:
                    # Align the msa
                    aligned_msa = repeat_align.realign_repeat(unaligned_msa)
                    if len(aligned_msa) > 1:
                        # Create a Repeat() class with the new msa
                        if 'repeat' in kwargs:
                            repeats.append(repeat.Repeat(aligned_msa,
                                                         **kwargs['repeat']))
                        else:
                            repeats.append(repeat.Repeat(aligned_msa))

            # Set begin coordinate for all repeats
            for i_repeat in repeats:
                self.repeat_in_sequence(i_repeat)

            return repeat_list.RepeatList(repeats)

        elif lHMM == []:
            LOG.debug("lHMM == []")
            return None

        elif denovo:
            if 'detection' in kwargs:
                predicted_repeats = repeat_detection_run.run_detector(
                    [self],
                    **kwargs['detection'])[0]
            else:
                predicted_repeats = repeat_detection_run.run_detector([self])[0]

            LOG.debug("predicted_repeats: {}".format(predicted_repeats))
            repeats = []

            for jTRD, jlTR in predicted_repeats.items():
                for iTR in jlTR:
                    if 'repeat' in kwargs:
                        iTR = repeat.Repeat(iTR.msa, begin=iTR.begin,
                                            **kwargs['repeat'])
                    else:
                        iTR = repeat.Repeat(iTR.msa, begin=iTR.begin)

                    # Consider only tandem repeats that have a repeat unit
                    # predicted to be at least one character long.
                    if iTR.l_effective > 0:

                        # Save l, n, MSA, TRD, scores, sequence_type, position
                        # in sequence of given type
                        iTR.TRD = jTRD

                        # Sanity check repeat and set begin coordinate for
                        # all repeats
                        if not self.repeat_in_sequence(iTR):
                            LOG.debug("The tandem repeat is not part of" \
                                      "the sequence. Detector: %s", iTR.TRD)
                            continue

                        repeats.append(iTR)

            return repeat_list.RepeatList(repeats)

        else:
            raise Exception("Either require denovo detection, or provide an",
                            "HMM")

    def get_repeatlist(self, tag):
        """ Retrieve `repeatlist` from this `sequence` instance.

        Retrieve `repeatlist` from this `sequence` instance. Access
        `repeatlist` as self.d_repeatlist[tag]

        Args:
            tag (str): A identifier for the repeat_list

        Returns:
            RepeatList: A repeat_list instance.
        """

        try:
            return self.d_repeatlist[tag]
        except ValueError:
            logging.error("RepeatList %s not in Sequence", tag)


    def set_repeatlist(self, repeatlist, tag):
        """ Add `repeatlist` as attribute to this `sequence` instance.

        Add `repeatlist` as attribute to this `sequence` instance. Access
        `repeatlist` as self.d_repeatlist[tag]

        Args:
            repeatlist (RepeatList): A repeat_list instance.
            tag (str): A identifier for the repeat_list
        """

        self.d_repeatlist[tag] = repeatlist

    def annotate(self, data, tag):

        self.d_annotations[tag] = data

    def get_annotation(self, tag):

        if tag in self.d_annotations:
            return self.d_annotations[tag]
        else:
            return []

    def repeat_in_sequence(self, my_repeat):
        """ Sanity check whether the `repeat` is part of this `sequence`. In
        case, calculate the position of the `repeat` within the `sequence`.

        If yes: Return True, set repeat.begin to corrected value if necessary.
        If no: Return False.
        Perform sanity check on sequences where all amino acids are or are
        converted to standard amino acids.

        Args:
            sequence (sequence): A sequence instance.

        Returns:
            bool: True if repeat is part of sequence, else false

        .. todo:: Decide whether save_original_msa is needed here.
        """

        repeat_sequence = repeat.get_repeat_sequence(my_repeat.msa_standard_aa)
        # The first letter in the sequence is counted as 1
        # (not 0, as in Python):
        starts = [m.start() + 1 for m in re.finditer(repeat_sequence,
                                                     self.seq_standard_aa)]

        if len(starts) != 0:  # Is the tandem repeat predicted correctly?
            if not hasattr(my_repeat, "begin") or my_repeat.begin not in starts:
                my_repeat.begin = starts[0]
            my_repeat.save_original_msa(self.seq)
            return True
        else:
            return False
